import numpy as np
from collections import deque

def fas_to_dic(file):
    
    file_content = open(file, 'r').readlines()
    seqs_list   = []
    
    for i in file_content:
        line = i.strip()
        if line: # just if file starts empty
            seqs_list.append(line) 
    
    keys = [] 
    values = []    
    i = 0
    while(">" in seqs_list[i]):
        keys.append(seqs_list[i])
        i += 1 
        JustOneValue = []

        while((">" in seqs_list[i]) == False):
            JustOneValue.append(seqs_list[i]) 
            i += 1

            if(i == len(seqs_list)):
                i -= 1
                break

        values.append("".join(JustOneValue).upper().replace(" ", ""))
        
    return dict(zip(keys, values))

# region: extended newick parser
class myNode:

    def __init__(self, name = None, 
                 left = None, 
                 right = None, 
                 ancestor = None,
                 index = None, 
                 hybrid = False,
                 node_label = '',
                 branch_length = 0.0,
                 support = 0.0,
                 gamma = 1.0, # probability
                 ):
        
        self.name = name
        self.left = left
        self.right = right
        self.ancestor = ancestor
        self.index = index
        self.branch_length = branch_length
        self.label = node_label
        self.hybrid = hybrid
        self.gamma = gamma
        self.support = support


        # it is true when node is hybrid and 
        # incoming hybrid nodes  are connected 
        # from their left to this node.
        # From this node sucessors are present.
        self.isChild = False # child from hybrid nodes

    def __repr__(self):
        base_str = f"Node(i = {self.index}"

        if self.name:
            base_str += f", n = {self.name}"

        if self.left:
            base_str += f", l = {self.left.index}"

        if  self.right and not isinstance( self.right, list):
                base_str += f", r = {self.right.index}"

        return base_str + ")"

def is_leaf(p):
    """
    Check if the node is a tree leaf node
    """
    if p.left is None and p.right is None:
        return True
    else:
        return False

def is_HybLeaf(p):
    """
    Check if the node is a hybrid leaf node
    """
    if p.hybrid and not p.isChild:
        return True
    else:
        return False


def make_node_info(p):
    """
    make node info.

    Ref: RichNewick.pdf (pp. 9-10)

    parameters:
    ----------
    p : myNode
        node to set info

    returns:
    -------
    ni : str
        node info string
    """
    la = f"{p.label}" if p.label else ''
    le = f"{p.branch_length}" if p.branch_length else ''
    su = f"{p.support}" if p.support else ''
    # no new info if 1.0
    pr = f"{p.gamma}" if p.gamma != 1.0 else ''

    ni = ''
    # combinations of 3
    if le and su and pr:
        ni += f":{le}:{su}:{pr}"

    # combinations of 2
    elif (not le) and su and pr:
        ni += f"::{su}:{pr}"

    elif le and (not su) and pr:
        ni += f":{le}::{pr}"

    elif le and su and (not pr):
        ni += f":{le}:{su}"
    
    # combinations of 1
    elif le and (not su) and (not pr):
        ni += f":{le}"

    elif (not le) and su and (not pr):
        ni += f"::{su}"

    elif (not le) and (not su) and pr:
        ni += f":::{pr}"

    else: # combination of 0
        pass

    if not is_leaf(p):
        ni = la + ni

    return ni


def writeT(p):

    if not p:
        return
    
    if is_leaf(p) or is_HybLeaf(p):
        # extract leaf node info
        node_info = make_node_info(p)
        return f"{p.name}{node_info}"
    
    else:
        ln = writeT(p.left)
        rn = writeT(p.right)
   
        node = f"{ln},{rn}" if rn else ln
        # extract node info only
        # when we reach the end 
        # of the stack
        node_info = make_node_info(p)
            
        return f"({node}){node_info}"

def parseTree(all_nodes, root = -1):

    w = all_nodes[root]

    if isinstance(w.right, list):
        n1 = w.left
        n2 = w.right[0]
        n3 = w.right[1]
        
        return f"({writeT(n1)},{writeT(n2)},{writeT(n3)});"
    
    else:
        n1 = w.left
        n2 = w.right

        return f"({writeT(n1)},{writeT(n2)});"

def parseBinTree(all_nodes, root = -1):

    w = all_nodes[root]
    n1 = w.left
    n2 = w.right

    return f"({writeT(n1)},{writeT(n2)});"

def renaming_tips(all_keys, tree):
    for n,k in enumerate(all_keys):
        # n,k
        spps = k.replace(">", "")
        tree = tree.replace(f"'{n}'", spps)

    return tree

def parse_and_rename(all_nodes, all_keys):
    nwk_str = parseTree(all_nodes)
    for n,k in enumerate(all_keys):
        nwk_str = nwk_str.replace(f"'t{n}'", k.replace(">", ""))
    return nwk_str

def get_int_nodes(n, T):
    int_nodes = [0]*(n - 2)
    k = 0
    for i in range(2*n - 2):

        if not T[i].name:
            int_nodes[k] = T[i].index
            k += 1

    return int_nodes

def get_edges(nodes_indx, n, T, get_root = False):
    
    E = [[0,0]]*(n - 3)
    k = 0; root = 0
    for u in nodes_indx:

        if isinstance(T[u].right, list):
            root = u
            continue
        
        E[k] = [u, T[u].ancestor.index]
        k += 1

    if get_root:
        return E, root
    
    else:
        return E
    
def tokenize(tree):
    """
    split into tokens based on the characters '(', ')', ',', ':', ';'
    Huelsenbeck, programming for biologists, pp. 124-125

    parameters:
    -----------
    tree : str
        newick tree string

    returns:
    --------
    tokens : list
        list of tokens
    """
    # tree = mytree
    tokens = []
    ns_size = len(tree)
    i = 0
    while i < ns_size:
        c = tree[i]
        if c in '(),:;':
            tokens.append(c)
            i += 1

        else:
            j = i
            tempStr = ''
            while c not in '(),:;':
                tempStr += c
                j += 1
                c = tree[j]

            i = j
            tokens.append(tempStr)

    return tokens
    
def set_float(s, k):
    try:
        return float(s)
    
    except ValueError:
        raise ValueError(f"Error: We expect a number at non-label node info. Check {s} at character {k}")
        
def next_nodeInfo(s):
    """
    This essentially checks if the next token
    is not a comma or a semicolon. If it neither
    of these, then it is a node info.
    """
    return True if (s != ',') or (s != ';') else False

def is_hybrid(s):
    """
    check if the string is a hybrid
    """
    if s.startswith('#'):
        return True
    else:
        return False

def set_nodeInfo(p, infoType, tk, k, hyb_nodes):
    """
    set node info based on the type of info
    
    0 : label
    1 : branch length
    2 : support
    3 : probability
    
    ref: RichNewick.pdf

    parameters:
    -----------
    p : myNode
        node to set
    infoType : int
        type of info to set
    tk : str
        token to set
    k : int
        character position in the string
    hyb_nodes : dict
        dictionary of hybrid nodes
        where the key is the id and the value
        is a list of nodes
    """

    if infoType == 0: # label
        # impossible to happen, but just in case
        assert not is_leaf(p), f"Error: We expect a label for a non-leaf node. Check {tk} at character {k}"
    
        p.label = tk
        p.hybrid = is_hybrid(tk)

        if p.hybrid:
            # update the hybrid nodes
            update_hyb_nodes(p, hyb_nodes, tk)

    elif infoType == 1:
        p.branch_length = set_float(tk, k)

    elif infoType == 2:
        p.support = set_float(tk, k)

    else:
        p.gamma = set_float(tk, k)


def has_children(p):
    if p.left is None or p.right is None:
        return False
    else:
        return True

def has_child(p):
    if p.left or p.right:
        return True
    else:
        return False    

def solve_polytomy(p, n, nodes):
    """
    solve polytomy by adding a new node y on the
    edge (p, gp) where gp is the grandparent of p.
    Draw:
          gp
          | (edge can be left or (any) right)
          y
         / \
        p   n3
       / \
      n1  n2 

    Distance between p and y equal to 0
    and y and grandparent equal to the previous
    distance between p and grandparent. 

    parameters:
    -----------
    p : myNode
        parent node 
    
    n : myNode
        new node

    nodes : list
        list of nodes in the tree

    returns:
    --------
    None        
    """

    # if p is not the root, 
    # then it has an ancestor
    gp = p.ancestor 
    
    # keep track of the side of p
    p_left = False
    if gp.left == p:
        p_left = True
    
    # make a new node y
    # make p and n children of y
    if p_left:
        # y -> gp
        y = myNode(left=p, right=n, ancestor=gp, 
                   branch_length=p.branch_length)
        # y <- gp, replace p with y
        gp.left = y
    else:
        # y -> gp
        y = myNode(left=n, right=p, ancestor=gp, 
                   branch_length=p.branch_length)
        # y <- gp, replace p with y
        # root case special case
        if isinstance(gp.right, list):
            if gp.right[0] == p:
                gp.right[0] = y
            else:
                gp.right[1] = y
        else:
            gp.right = y

    # add y to the list of nodes
    nodes += [y]

    # if it is polytomy, then most likely this is already 
    # 0, but we set it to 0 anyway
    p.branch_length = 0.
    p.ancestor = y
    n.ancestor = y


def set_left_or_right(p, n, root, nodes):
    """
    set the left or right child of a node

    if p is the root, then the right child
    is a list of the current right child and the new node n

    parameters:
    -----------
    p : myNode
        parent node

    n : myNode
        new node

    root : myNode
        root node
    
    nodes : list
        list of nodes in the tree. This is used to
        to add an intermediate node in case of a polytomy
        see `solve_polytomy`

    returns:
    --------
    None

    """
    # set left child if it is None
    # as first option
    if p.left is None:
        p.left = n
        return
    
    # set right child if it is None
    # as second option
    if p.right is None:
        p.right = n
        return
    
    # if left and right children are not None
    # then the right child is a list of the current
    # right child and the new node n if p is the root
    if p == root:
        p.right = [p.right, n]
    else:
        # otherwise, we have a polytomy.
        # if p is not the root, then it has an ancestor
        # and it is a polytomy. We solve the polytomy
        solve_polytomy(p, n, nodes)


def build_up_nodes(tokens):

    hyb_nodes = {}
    # ensures O(1) appending 
    nodes = deque()
    root = None
    p = None

    readingNodeInfo = False
    # Info Type: 0 for label, 
    # 1 for edge, 2 for support, 
    # 3 for probability
    infoType = 0 

    # k tracks the position in the string
    k = 0
    for i, tk in enumerate(tokens):

        k += len(tk)

        if tk == "(":
            # internal node
            n = myNode()
            nodes.append(n)
            if p is None:
                root = n
            
            else:
                # connect the node to the parent
                n.ancestor = p
                # set new node the left or right child
                set_left_or_right(p, n, root, nodes)
            
            # let the current node 
            # be the parent
            p = n

        elif tk == "," or tk == ")":
            # move down a node
            p = p.ancestor

            if tk == ")":
                # check if the node 
                # is not empty
                if not has_child(p):
                    raise ValueError(f"Error: We expect at least a child per node. Check character {k}")
                
                # check if there info for the 
                # internal node coming
                if next_nodeInfo(tokens[i+1]):
                    # start reading internal node info.
                    readingNodeInfo = True
                    # we iterate all ':' tokens
                    # to find the type of info
                    # we start from 0
                    infoType = 0 
            
            # it means we reached ',' and we stop
            # reading node info
            else:
                readingNodeInfo = False
                infoType = 0

        elif tk == ";":
            # end of tree
            if p != root:
                raise ValueError("Error: We expect to finish at the root node")

        else:
            if readingNodeInfo:
                if tk == ":":
                    # this indicates that there 
                    # is info on the next token.

                    # Notice that if p is leaf node, 
                    # the next token is ":" and then
                    # it will start reading from length
                    # as infoType will be 1:
                    infoType += 1
                    # then we go to the next token
                    continue

                # reading node info
                if 0 <= infoType <= 3:
                    set_nodeInfo(p, infoType, tk, k, hyb_nodes)

            else:
                # leaf node
                leaf_name = tk.strip("''")
                n = myNode(name=leaf_name, hybrid=is_hybrid(leaf_name))
                
                # update set of nodes
                nodes.append(n)
                # update set of hybrid nodes
                if n.hybrid:
                    update_hyb_nodes(n, hyb_nodes, leaf_name)

                # connect the leaf node to the parent
                n.ancestor = p
                # set new node to left or right
                set_left_or_right(p, n, root, nodes)

                # let the current node 
                # be the parent
                p = n

                # check if there is leaf node info coming
                if next_nodeInfo(tokens[i+1]):
                    readingNodeInfo = True
                    # Iterations will decide 
                    # the type of info
                    infoType = 0

    if hyb_nodes:
        # process the hybrid nodes
        connect_hyb_nodes(hyb_nodes)

    return list(nodes), hyb_nodes, root

def connect_hyb_nodes(hyb_nodes):
    """
    connect the hybrid nodes to one child,
    the one that is not a leaf node (i.e., it
    has no successors). Child node will have
    multiple ancestors. The first ancestor will be
    tree base node. The rest of the ancestors
    and the rest of ancestors will be hybrid leaf nodes.

    The left of hybrid leaf nodes will
    be connected to the child node.

    parameters:
    -----------
    hyb_nodes : dict
        dictionary of hybrid nodes
        where the key is the id and the value
        is a list of nodes
        e.g. {#H1: [n1, n2, n3]}
        where n1, n2, n3 are hybrid nodes,
        with one of them being a child node
        and the rest are leaf nodes.

    returns:
    -------
        None
    """
    for k, nodes in hyb_nodes.items():    
        # recieving all the leaf nodes
        # this node has successors
        out_node = None # child node
        # injecting the leaf nodes
        in_nodes = deque()
        
        for n in nodes:
            if is_leaf(n):
                in_nodes.append(n)
            else:
                out_node = n

        assert len(in_nodes) == len(nodes) - 1, f"Error: We found {len(in_nodes)} leaf nodes representing {k} when it should be {len(nodes) - 1} (i.e. number of {k} nodes - 1)"
        assert out_node, f"Error: We found no {k} nodes with successors"

        out_node.ancestor = [out_node.ancestor] + list(in_nodes)
        out_node.isChild = True
        
        # leaf nodes are empty
        # and they have no successors
        for n in in_nodes:
            n.left = out_node


def update_hyb_nodes(p, hyb_nodes, id):
    """
    update the hybrid nodes

    parameters:
    -----------
    p : myNode
        node to set

    hyb_nodes : dict
        dictionary of hybrid nodes
        where the key is the id and the value
        is a list of nodes

    id : str
        id of the hybrid node e.g. #H1, #H2, etc. 
        if p is a internal node then
        the id is the label of the node
        if p is a leaf node then 
        the id is the name of the node

    returns:
    -------
        None
    """
    if id not in hyb_nodes:
        hyb_nodes[id] = deque([p])

    else:
        hyb_nodes[id].append(p) # O(1)

def parseNewickTree(mstr):
    # mstr = mytree
    tokens = tokenize(mstr)
    # print(tokens)

    nodes, _, root = build_up_nodes(tokens)
    # print(nodes[0].right)
    for i, n in enumerate(nodes):
        n.index = i

        if n.ancestor is None:
            root = n

    return nodes, root

def parseNewickNet(mstr):
    # mstr = mytree
    tokens = tokenize(mstr)
    # print(tokens)

    nodes, hyb_nodes, root = build_up_nodes(tokens)
    # print(nodes[0].right)
    for i, n in enumerate(nodes):
        n.index = i

        if n.ancestor is None:
            root = n

    return nodes, hyb_nodes, root
# endregion: extended newick parser

# region: vcv matrix
def get_nodes(n):
    # 3*(2n-2) = O(n)
    out = []
    if isinstance(n, list):
        for ni in n:
            out.append(ni)
    else:
        if n is not None:
            out.append(n)

    return out

def get_adj(nodes):
    adj = {}
    for n in nodes:
        adj[n] = []
        adj[n] += get_nodes(n.left)
        adj[n] += get_nodes(n.right)
        adj[n] += get_nodes(n.ancestor)

    return adj

def dfs_distances(adj, start, C):
    # O(n) since it's a tree
    S = [(start, -1, 0)]   # (node, parent, distance)
    dist = {}

    while S:
        node, parent, d = S.pop()
        dist[node] = d

        for u in adj[node]:
            if u != parent:
                S.append((u, node, d + C[(node, u)]))

    return dist

def get_weight(v, u):
    # if isinstance(v.ancestor, list):
    if u.left == v or u.right == v:
        return v.branch_length
    else:        
        return u.branch_length

def get_C(nodes, adj):
    """
    get edge weights
    O(n) since it's a tree
    """
    C = {} # O(n)
    for v in nodes:
        for u in adj[v]:
            w = get_weight(v, u)
            C[(v, u)] = w
            C[(u, v)] = w

    return C

def get_APSP(leaves, adj, C):
    n = len(leaves)
    D = np.zeros((n, n))

    # O(n^2)
    for i in range(n - 1): # O(n)
        u = leaves[i]
        tmp_dist = dfs_distances(adj, u, C) # O(n)
        for j in range(i + 1, n):
            v = leaves[j]
            D[i, j] = tmp_dist[v]
            D[j, i] = tmp_dist[v]

    return D

def get_BM(D, root, leaves, adj, C):
    n = len(leaves)
    dist_root = dfs_distances(adj, root, C)
    D_r = np.zeros(n)
    # O(n)
    for i in range(n):
        D_r[i] = dist_root[leaves[i]]

    # O(n^2)
    BM_VCV = np.zeros((n, n))
    for i in range(n):
        BM_VCV[i, :] = (D_r[i] + D_r - D[i, :])/2

    return BM_VCV

def get_cov_mat(nodes, root, process = "BM",
                 sigma2 = 1, alpha = 1):
    # target nodes: leaves 
    leaves = [n for n in nodes if is_leaf(n)]
    names = [n.name for n in leaves]

    adj = get_adj(nodes) # O(n)
    C = get_C(nodes, adj) # O(n)

    # O(n^2), it transverses the tree n times, and
    # each traversal is O(n)
    D = get_APSP(leaves, adj, C)
    # print(D)
    if process == "BM":
        # makes for each row i it makes n calculations from D.
        # since there n rows, the time complexity is O(n^2)
        BM_D = get_BM(D, root, leaves, adj, C) # O(n^2)
        return sigma2*BM_D, names
    
    elif process == "OU":
        c1 = 2*alpha
        return (sigma2/c1)*np.exp(-c1*D), names
    
    else:
        raise ValueError("process must be either 'BM' or 'OU'")

def get_vcv(mytree, process = "BM", sigma2 = 1, alpha = 0.5):
    """
    get the variance-covariance matrix from
    a newick tree string.

    O(n^2) time complexity, where n is the number of tips in the tree.

    parameters:
    -----------
    mytree : str
        newick tree string

    process : str
        "BM" for Brownian motion
        "OU" for Ornstein-Uhlenbeck

    sigma2 : float
        evolutionary rate

    alpha : float
        selection parameter, only used for OU process

    returns:
    --------
    vcv : list
        variance-covariance matrix without
        being multiplied by the evolutionary rate

    name_list : list
        ordered list of tip names
        used to build the vcv matrix
    """

    nodes, root = parseNewickTree(mytree)
    assert not isinstance(root.right, list), "Unrooted tree not supported"

    vcv, names = get_cov_mat(nodes, root, process = process, sigma2 = sigma2, alpha = alpha)
    # return the variance-covariance matrix
    return vcv, names

# endregion: vcv matrix

# # testings of code
# mytree1 = "((t2:3,t3:2):1,t4,t1:4);"
# mytree2 = "(t4,(t2:3,t3:2):1,t1:4);"
# mytree3 = "(t4,t1:4,(t2:3,t3:2):1);"
# mytree4 = "(t4,t1:4,(t2:3,t3:2,t5:1):1);"
# for i in [mytree1, mytree2, mytree3, mytree4]:
#     tmp_nodes, tmp_root = parseNewickTree(i)
#     print(parseTree(tmp_nodes, root = tmp_root.index))

# mytree = "(t4,t1:4,(t2:3):1);"
# nodes, root = parseNewickTree(mytree)
# print(root.right)
# print(parseTree(nodes, root = root.index))

# mytree = "((t4,t1:4),(t2:3):1);"
# vcv, names = get_vcv(mytree)

# import numpy as np
# np.array(vcv)


import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import cholesky

def make_circles(rho, num_samples, seed, test_num = 0.5):

    np.random.seed(seed)

    # # Step 1: Define correlation structure for observations
    # num_samples = 500
    # rho = 0.5  # Strength of correlation between consecutive observations

    # Construct correlation matrix (exponential decay)
    V = np.array([[rho**abs(i-j) for j in range(num_samples)] for i in range(num_samples)])

    # Step 2: Generate independent Gaussian features
    num_features = 2
    X_uncorrelated = np.random.randn(num_samples, num_features)

    # Step 3: Apply Cholesky decomposition to correlate observations
    L = cholesky(V)
    X = L @ X_uncorrelated  # Now observations are correlated

    # Step 4: Define a **circular non-linearity** (donut shape)
    radius = .01  # Set the radius for the decision boundary
    decision_boundary = np.sqrt(X[:, 0]**2 + X[:, 1]**2)  # Compute distance from origin
    y = ((decision_boundary > radius) & (decision_boundary < radius + 1)).astype(int)  # Define labels

    # Step 5: Shuffle the data and split into training and testing sets
    all_index = list(range(X.shape[0]))
    np.random.shuffle(all_index)
    
    X = X[all_index]
    y = y[all_index]
    V = V[all_index, :][:, all_index]


    Indx = int(test_num*X.shape[0])
    X_train = X[:Indx]
    X_test = X[Indx:]
    y_train = y[:Indx]
    y_test = y[Indx:]
    V_train = V[:Indx, :Indx]
    V_test = V[Indx:, Indx:]

    return X_train, X_test, y_train, y_test, V_train, V_test

def plot_dataset(X, y, axes):
    plt.plot(X[:, 0][y==0], X[:, 1][y==0], "bs")
    plt.plot(X[:, 0][y==1], X[:, 1][y==1], "g^")
    plt.axis(axes)
    plt.grid(True, which='major', linestyle='--', linewidth='0.5')
    plt.xlabel(r"$x_1$", fontsize=20)
    plt.ylabel(r"$x_2$", fontsize=20, rotation=0)

def plot_predictions(clf, axes):
    x0s = np.linspace(axes[0], axes[1], 100)
    x1s = np.linspace(axes[2], axes[3], 100)
    x0, x1 = np.meshgrid(x0s, x1s)
    X = np.c_[x0.ravel(), x1.ravel()]
    y_pred = clf.predict(X).reshape(x0.shape)
    plt.contourf(x0, x1, y_pred, cmap=plt.cm.bwr , alpha=0.3)


from krgee import Logistic, k_fold_cv_random_vcv, k_fold_cv_random_vcv_v2
import matplotlib.pyplot as plt


(X_train, X_test, 
 y_train, y_test, 
 vcv_train, vcv_test) = make_circles(rho=0.5, 
                                 num_samples=500, 
                                 seed=12038, 
                                 test_num=0.5)


lkrr = Logistic(get_corr=False, tol=1e-4, 
               max_iter=400, kernel='rbf', 
               verbose=True)

params = {
    'lambda' : np.logspace(-5, 3, 200, base=2),
    'gamma' : np.logspace(-5, 3, 200,  base=2),
}

best_params = k_fold_cv_random_vcv_v2(X_train, y_train, vcv_train, 
                                   lkrr, params,
                                   folds=3, sample=100, 
                                   verbose=True, seed=12038)
print('Best params: ', best_params)
lkrr.set_params(**best_params)
lkrr.fit(X_train, y_train, vcv_train) 
print(lkrr.score(X_test, y_test))


from sklearn.linear_model import LogisticRegression
log_reg = LogisticRegression()
log_reg.fit(X_train, y_train)
log_reg_loss = 1 -  log_reg.score(X_test, y_test)

print( 'Loss: ', round(log_reg_loss, 2) )

tmp_axes =  [-4, 4, -4, 4]
fig, axes = plt.subplots(1, 2, figsize=(10, 4))
plt.sca(axes[0])
plot_predictions(log_reg, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)

plt.sca(axes[1])
plot_predictions(lkrr, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)
plt.tight_layout()
# plt.savefig('classification_comparisons.png', dpi=300)







import numpy as np
import random
import numpy as np
import glob
# import matplotlib.pyplot as plt

# seed for reproducibility
np.random.seed(12038)
random.seed(12038)

def split_data(X,y, num_test, seed = 123):
    random.seed(seed)
    n,_ = X.shape

    test_idx = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    return X_train, X_test, y_train, y_test


def split_data_vcv(X,y,vcv, num_test, seed = 123):
    random.seed(seed)
    n,_ = X.shape

    test_idx = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]
    vcv_train, vcv_test = vcv[train_idx,:][:,train_idx], vcv[test_idx,:][:,test_idx]

    return X_train, X_test, y_train, y_test, vcv_train, vcv_test

def R2(y, r):
    """
    Coefficient of determination
    r = y - y_pred
    """

    # r = y - y_pred
    u = r.T @ r
    v = y.T @ y - (len(y) * np.mean(y)**2)
    return 1 - (u/v)

def rmse(r):
    """
    Root Mean Squared Error
    r = y - y_pred
    """
    # r = y - y_pred
    return np.sqrt( np.mean( r**2 ) )

def distance_matrix(a, b):
    """
    l2 norm squared matrix
    """
    return np.linalg.norm(a[:, None, :] - b[None, :, :], axis=-1)**2

def RBF_kernel(a, b, gamma):
    """
    Radial Basis Function
    """

    tmp_rbf = -gamma * distance_matrix(a, b)
    np.exp(tmp_rbf, tmp_rbf) # RBF kernel. Inplace exponentiation
    return tmp_rbf

def linear_kernel(a, b, c):
    """
    Linear Kernel
    """
    XXt = a.dot(b.T)
    C = c * np.ones(XXt.shape)

    return XXt + C

def P_mat_simple(vcv):
    """
    get the square root of the inverse of the
    """

    # Kr = np.diag(1/np.sqrt(np.diag(vcv)))
    # vcv = Kr @ vcv @ Kr

    if isinstance(vcv, type(None)):
        return None

    L,Q = np.linalg.eig( vcv )
    P = Q @ np.diag( L**(-1/2) ) @ Q.T

    return P

def _scaler(ref, dat, use_sd = False):
    """
    Center and scale data
    """

    u = np.mean(ref, axis=0)
    centered = dat - u

    if use_sd:
        sd = np.std(ref, axis=0)
        sted = centered / sd

        return sted
    else:

        return centered

def scale_weighted_data(X, vcv, use_sd = False):
    """
    scale data by the square root of the inverse of the
    covariance matrix

    X: np.array
        data

    vcv: np.array
        covariance matrix
    """

    P = P_mat_simple(vcv)
    X_w_uc = P @ X
    return _scaler(X_w_uc, X_w_uc, use_sd)

def k_fold_cv(X, y, model, num_folds):
    """
    k-fold cross-validation
    """

    n, p = X.shape
    fold_size = n // num_folds

    all_errors = []
    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]
    
        model.fit(X_train, y_train)
        tmp_err = model.score(X_test, y_test, metric='rmse')
        all_errors.append(tmp_err)

    return np.mean(all_errors)

def k_fold_cv_random(X, y, model, params, folds=3, sample=500, verbose=True, seed=123):
    """
    Random search for hyperparameter tuning using k-fold cross-validation
    """
    np.random.seed(seed=seed)

    # make random choice from the grid of hyperparameters
    all_params = params.keys()
    tested_params = np.zeros((sample, len(all_params)))

    for n, k in enumerate(all_params):
        tested_params[:, n] = np.random.choice(params[k], sample)

    if verbose:
        # check tested_params are unique
        tested_params = np.unique(tested_params, axis=0)
        print("Number of unique hyperparameters: ", tested_params.shape[0])

    # shuffle the data
    n = X.shape[0]
    idx = np.arange(n)
    np.random.shuffle(idx)
    X = X[idx]
    y = y[idx]

    all_errors = []
    for vec in tested_params:
        tmp_params = dict(zip(all_params, vec))
        model.set_params(**tmp_params)
        tmp_err = k_fold_cv(X, y, model, folds)
        all_errors.append([tmp_params, tmp_err])

        if verbose:
            print("CV score: %s, Parameters %s" % (tmp_err, tmp_params))

    # take the best hyperparameters
    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]

    if verbose:
        print("Best CV score: ", best_[1])

    return best_[0]


class PGLS:
    def __init__(self, fit_intercept=False, weighted=False) -> None:
        self.fit_intercept = fit_intercept
        self.weighted = weighted
        self.intercept = 0
        self.beta = np.array([])

    def fit(self, X, y, vcv):
        n, p = X.shape

        if self.weighted:
            tmp_beta = np.linalg.solve(X.T @ X, X.T @ y)

        if self.fit_intercept:
            self.intercept = np.mean(y - X @ tmp_beta)

        self.beta = np.hstack((self.intercept, tmp_beta)) if self.fit_intercept else tmp_beta

    def predict(self, X):
        n, p = X.shape

        if self.fit_intercept:
            X = np.hstack((np.ones((n, 1)), X))

        return X @ self.beta

    def score(self, X_test, y_test, metric='rmse'):
        y_pred = self.predict(X_test)

        if metric == 'rmse':
            return np.sqrt(np.mean((y_pred - y_test) ** 2))
        else:
            u = ((y_test - y_pred) ** 2).sum()
            v = ((y_test - y_test.mean()) ** 2).sum()

            return 1 - (u / v)


class WKRR:
    def __init__(self, kernel='rbf', fit_intercept=True, check_cov=False) -> None:
        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        self.fit_intercept = fit_intercept
        self.check_cov = check_cov
        self.intercept = 0
        self.alpha = np.array([])
        self.X = np.array([])
        self.chol = False

    def set_params(self, **params):
        if self.kernel == 'rbf':
            self.params['gamma'] = params['gamma']

        else:
            self.params['c'] = params['c']

        self.params['lambda'] = params['lambda']

    def get_params(self):
        return self.params

    def P_mat(self, vcv):
        if isinstance(vcv, type(None)):
            return None

        if self.check_cov:
            self.assert_COV_sym(vcv)

        if self.chol:
            C = np.linalg.cholesky(vcv)
            P = np.linalg.inv(C)

        else:
            L, Q = np.linalg.eig(vcv)
            P = Q @ np.diag(L ** (-1 / 2)) @ Q.T

        if self.check_cov:
            self.assert_COV_decom(P, vcv)

        return P

    def assert_COV_sym(self, vcv, tol=1e-8):
        assert np.all(np.abs(vcv - vcv.T) < tol), 'not symmetric matrix'

    def assert_COV_decom(self, P, vcv):
        assert np.all(np.round(P.T @ P, 2) == np.round(np.linalg.inv(vcv), 2)), "P.T @ P != vcv^-1"

    def fit(self, X, y, vcv=None):
        self.X = X
        P = self.P_mat(vcv)

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])
        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, y, self.params['lambda'], P)

        if self.fit_intercept:
            self.intercept = np.mean(y - K_train @ self.alpha)

    def predict(self, X_test):
        assert len(self.alpha) > 0, "The model needs to be fitted first"

        if self.kernel == 'rbf':
            K_test = RBF_kernel(X_test, self.X, self.params['gamma'])
        else:
            K_test = linear_kernel(X_test, self.X, self.params['c'])

        return K_test @ self.alpha + self.intercept

    def score(self, X_test, y_test, vcv_test, metric='rmse'):
        y_pred = self.predict(X_test)

        if isinstance(vcv_test, type(None)):
            P = np.eye(X_test.shape[0])
        else:
            P = self.P_mat(vcv_test)

        werr = P @ (y_pred - y_test)  # weighted residuals

        if metric == 'rmse':
            return rmse(werr)
        else:
            Py = P @ y_test  # weighted targets
            return R2(Py, werr)

    def opt_alpha(self, K, y, reg_lam=None, P=None):

        n, _ = self.X.shape
        I = np.eye(K.shape[0])
        nlI = n * reg_lam * I

        if isinstance(P, type(None)):
            return np.linalg.solve(K + nlI, y)
        
        else:
            return P @ np.linalg.solve(P @ K @ P + nlI, P @ y)


class KRR(WKRR):

    def __init__(self, kernel='rbf', fit_intercept=True) -> None:
        super().__init__(kernel, fit_intercept)
        self.intercept = 0
        self.alpha = np.array([])
        self.X = np.array([])

    def fit(self, X, y):
        self.X = X
        self.y = y

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])
            
        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, self.y, self.params['lambda'])

        if self.fit_intercept:
            self.intercept = np.mean(y - K_train @ self.alpha)

    def opt_alpha(self, K, y, reg_lam=None):
        
        n, _ = self.X.shape
        I = np.eye(K.shape[0])
        nlI = n * reg_lam * I

        return np.linalg.solve(K + nlI, y)
        

    def score(self, X_test, y_test, metric='rmse'):
        y_pred = self.predict(X_test)

        if metric == 'rmse':
            return np.sqrt(np.mean((y_pred - y_test) ** 2))
        else:
            u = ((y_test - y_pred) ** 2).sum()
            v = ((y_test - y_test.mean()) ** 2).sum()

            return 1 - (u / v)


def myfunc(x, b=0, type='sin'):
    if type == 'sin':
        return np.sin( np.sum(x, axis=1) * 2).ravel()
    
    else:
        return np.sum(x, axis=1) ** b

def sim_data(vcv, mean_vector, b, 
             on_weighted=False, 
             add_noise = False, 
             noise_var = 1,
             n_var = 3, 
             extra_weights = [1,1,1], 
             type='pol'):
    """
    Simulate data for n_var independent variables

    Parameters
    ----------
    vcv: np.array
        covariance matrix
    
    mean_vector: np.array
        mean vector
    
    b: int
        power of the polynomial function

    on_weighted: bool
        whether to use weighted data or not to obtain the response variable.
        the weights are obtained using the 
        squared root of the inverse of 
        the covariance matrix

    add_noise: bool
        whether to add noise to the response variable

    noise_var: float
        variance of the noise

    extra_weights: list
        weights for the independent variables

    type: str
        type of the function to simulate the data
        if 'sin' then a sine function is used over the
        sum of the independent variables
        if 'pol' then a polynomial function is used over the
        sum of the independent variables
    
    Returns
    -------
    X_w_uc: np.array
        weighted independent variables (uncentered)
    
    y_w_uc: np.array
        weighted response variable (uncentered)
    """
    n_var = 3

    assert len(extra_weights) == n_var, "extra weights must have the same length as the number of predictors"

    X_uw_uc = np.zeros((vcv.shape[0], n_var))
    for j in range(n_var):
        X_uw_uc[:,j] = np.random.multivariate_normal(mean=mean_vector, cov=vcv)*extra_weights[j]


    P = P_mat_simple(vcv)
    noise = np.random.normal(0, noise_var, X_uw_uc.shape[0])

    if on_weighted:
        X_w_uc = P @ X_uw_uc
        y_w_uc = myfunc(X_w_uc, b=b, type=type).ravel()

        if add_noise:
            y_w_uc += noise

        # plt.scatter(np.sum(X_w_uc, 1), y_w_uc, color='blue', alpha=0.5)
    else:
        y_uw_uc = myfunc(X_uw_uc, b=b, type=type).ravel()

        if add_noise:
            y_uw_uc += noise        

        X_w_uc = P @ X_uw_uc
        y_w_uc = P @ y_uw_uc

    return X_w_uc, y_w_uc


def test_vcv(vcv, b=4, on_weighted=True, sample=50, add_noise=False, noise_var=1):

    n = vcv.shape[0]
    mean_vector = np.zeros(n)
    num_test = round(0.40 * n)

    X_w_uc, y_w_uc = sim_data(vcv, mean_vector, b=b, 
                              on_weighted=on_weighted,
                              add_noise=add_noise,
                              noise_var=noise_var)

    X_wc = _scaler(X_w_uc, X_w_uc, use_sd= False)
    y_wc = _scaler(y_w_uc, y_w_uc, use_sd= False)

    (X_wc_train, X_wc_test,
     y_wc_train, y_wc_test,
     vcv_train , vcv_test) = split_data_vcv(X_wc, y_wc, vcv, num_test=num_test, seed=12038)

    rbf_param = k_fold_cv_random(X_wc_train, y_wc_train, krr_rbf, rbf_grid_params, verbose=False, folds=3, sample=sample)
    linear_param = k_fold_cv_random(X_wc_train, y_wc_train, krr_linear, linear_grid_params, verbose=False, folds=3, sample=sample)

    krr_rbf.set_params(**rbf_param)
    krr_linear.set_params(**linear_param)

    krr_rbf.fit(X_wc_train, y_wc_train)
    krr_linear.fit(X_wc_train, y_wc_train)
    pgls_model.fit(X_wc_train, y_wc_train, vcv_train)

    rmse_wc_krr_rbf = krr_rbf.score(X_wc_test, y_wc_test, metric='rmse')
    rmse_wc_krr_linear = krr_linear.score(X_wc_test, y_wc_test, metric='rmse')
    rmse_wc_pgls = pgls_model.score(X_wc_test, y_wc_test, metric='rmse')

    error_sim = [b, rmse_wc_krr_rbf, rmse_wc_krr_linear, rmse_wc_pgls]
    print(error_sim)
    return error_sim

def permutation_importance(X, y, model, num_test = 100, seed = 12038, iterations = 1000):

    np.random.seed(seed)

    n,p = X.shape

    X_train, X_test, y_train,y_test = split_data(X, y, num_test, seed = seed)
    model.fit(X_train, y_train)
    error_orig = model.score(X_test, y_test)


    out = np.zeros((iterations, p))

    for i in range(iterations):

        for j in range(X_test.shape[1]):
            # Create a copy of X_test
            X_test_copy = X_test.copy()

            # Scramble the values of the given predictor
            X_test_copy[:,j] = np.random.permutation(X_test_copy[:,j])
                        
            # Calculate the new RMSE
            error_perm = model.score(X_test_copy, y_test)

            out[i,j] = error_perm - error_orig

    return out

def pdp(self, feature, integer = False, quantiles = [0,1], sample = 70):
    """
    partial depedence plot values.

    Parameters
    ----------
    feature : int

    Returns
    -------
    PDP of the transformed data
    """
    if not len(self.P_inv):
        self.P_inv = np.linalg.inv(self.P)

    Xq = self.sample_feature(feature, quantiles, sample, integer)
    
    pdp_values = []
    for n in Xq:
        # copy original values
        X_tmp = self.X.copy()
        # make original values with 
        # modified feature column
        X_tmp[:,feature] = n
        # weight the whole dataset as
        # the model learned from the 
        # weighted data
        X_tmp_test = self.P @ X_tmp

        pdp_values.append(
            np.mean(
                # make rows unweighted
                # (i.e., again 'correlated')
                self.P_inv @ self.model.predict(X_tmp_test)
            )
        )

    out = np.hstack((
        Xq.reshape(-1,1), 
        np.array(pdp_values).reshape(-1,1)
    ))
    return out

rbf_grid_params = {
    'lambda': np.logspace(-10, 5, 100, dtype=float, base=2),
    'gamma': np.logspace(-10, 5, 100, dtype=float, base=2),
}

linear_grid_params = {
    'c': np.logspace(-10, 5, 100, dtype=float, base=2),
    'lambda': np.logspace(-10, 5, 100, dtype=float, base=2)
}

krr_rbf    = KRR(kernel='rbf', fit_intercept=not True)
krr_linear = KRR(kernel='linear', fit_intercept=not True)
pgls_model = PGLS(fit_intercept=not True, weighted=True)

path_trees = '/Users/ulises/Desktop/ABL/software/phylokrr/data/r_coal_sim_500spps_100trees'
vcvs = glob.glob(path_trees + '/*.csv')

vcv = np.loadtxt(vcvs[0], delimiter=',')
b = 2
on_weighted = True
add_noise = True
noise_var = 1
sample = 100


n = vcv.shape[0]
mean_vector = np.zeros(n)
num_test = round(0.40 * n)

X_w_uc, y_w_uc = sim_data(vcv, mean_vector, b=b, 
                            on_weighted=on_weighted,
                            add_noise=add_noise,
                            noise_var=noise_var,
                            extra_weights=[3,2,1]
                            )

X_wc = _scaler(X_w_uc, X_w_uc, use_sd= False)
y_wc = _scaler(y_w_uc, y_w_uc, use_sd= False)

(X_wc_train, X_wc_test,
    y_wc_train, y_wc_test,
    vcv_train , vcv_test) = split_data_vcv(X_wc, y_wc, vcv, num_test=num_test, seed=12038)

rbf_param = k_fold_cv_random(X_wc_train, y_wc_train, krr_rbf, rbf_grid_params, verbose=False, folds=3, sample=sample)
linear_param = k_fold_cv_random(X_wc_train, y_wc_train, krr_linear, linear_grid_params, verbose=False, folds=3, sample=sample)

krr_rbf.set_params(**rbf_param)
krr_rbf.fit(X_wc_train, y_wc_train)
# krr_linear.fit(X_wc_train, y_wc_train)

def sample_feature(X, feature, quantiles, sample, integer = False):
    """
    sample unweighted column
    """
    # X = X0
    # feature =0
    Xpdp = X[:,feature]

    if integer:
        return np.sort(np.unique(Xpdp))
    
    q = np.linspace(
        start = quantiles[0],
        stop  = quantiles[1],
        num   = sample
    )

    # sample from the original feature space
    Xq = np.quantile(Xpdp, q = q)

    return Xq    

def pdp_1d(X, feature, model, quantiles = [0,1], sample = 70, integer = False):
    """
    1D partial dependence plot
    """

    Xq = sample_feature(X, feature, quantiles, sample, integer)
    pdp_values = []
    for n in Xq:
        # copy original values
        X_tmp = X.copy()
        # make original values with 
        # modified feature column
        X_tmp[:,feature] = n

        pdp_values.append(
            np.mean(
                model.predict(X_tmp)
            )
        )

    out = np.hstack((
        Xq.reshape(-1,1), 
        np.array(pdp_values).reshape(-1,1)
        ))
    
    return out


import matplotlib.pyplot as plt


def pdp_2d(X, f1_idx, f2_idx, model, quantiles, sample, integer):
    """
    2D partial dependence plot
    """

    Xq1 = sample_feature(X, f1_idx, quantiles, sample, integer[0])
    Xq2 = sample_feature(X, f2_idx, quantiles, sample, integer[1])

    n_xq1 = len(Xq1)
    n_xq2 = len(Xq2)

    X1, X2 = np.meshgrid(Xq1, Xq2)

    Y = np.zeros((n_xq1, n_xq2))
    for i in range(n_xq1):
        for j in range(n_xq2):
            # j,i = 0,0
            X_tmp = X.copy()
            X_tmp[:,f1_idx] = X1[i,j]
            X_tmp[:,f2_idx] = X2[i,j]

            Y[i,j] = np.mean( model.predict(X_tmp) )

    return X1, X2, Y

# 2d partial dependence plot
sample = 10
f1_idx = 2
f2_idx = 1
X = X_wc
quantiles = [0,1]
integer = [False, False]
model = krr_rbf


X1, X2, Y = pdp_2d(X, f1_idx, f2_idx, model, quantiles, sample, integer)

fig,ax=plt.subplots(1,1)
cp = ax.contourf(X1, X2, Y)
fig.colorbar(cp) # Add a colorbar to a plot
ax.set_title('2D partial dependence plot')
ax.set_xlabel('feature 1')
ax.set_ylabel('feature 2')



# permutation importance
out = permutation_importance(X_wc, y_wc,krr_rbf, num_test = 100, iterations=5)

# plot bar chart
import matplotlib.pyplot as plt
plt.bar(range(3), np.mean(out, axis=0))
plt.errorbar(range(3), np.mean(out, axis=0), np.std(out, axis=0), fmt='.', color='black')
plt.xticks(range(3), ['feature 1', 'feature 2', 'feature 3'])
plt.ylabel('RMSE increase')


pdp_1d_out = pdp_1d(X_wc, 0, krr_rbf)

# make a line plot
plt.plot(pdp_1d_out[:,0], pdp_1d_out[:,1])







import random
import numpy as np

from phylokrr.utils import P_inv_simple, P_mat_simple
from phylokrr.datasets import load_vcv

np.random.seed(12037)

vcv = load_vcv()
mean_vector = np.zeros(vcv.shape[0])

X_uw_uc = np.random.multivariate_normal(mean_vector + 4, vcv, 1).reshape(-1,1)
# Non-linear response variable (sine curve)
y_uw_uc = np.sin(X_uw_uc*1.5).reshape(-1,1)

# # Add noise to the response variable
# y_w_uc[::10] += 4 * (0.5 - np.random.rand(X_w_uc.shape[0] // 10))

# Add noise to the response variable
e = np.random.multivariate_normal(mean_vector, vcv, 1).reshape(-1,1)
# y_uw_uc[::2] += 2 * (0.5 + e[::2])
y_uw_uc += e

# we can attempt to unweight to the original space
# with the square root of the covariance matrix
# P_inv = P_inv_simple(vcv)
P = P_mat_simple(vcv)

X_w_uc, y_w_uc = P @ X_uw_uc, P @ y_uw_uc

import matplotlib.pyplot as plt
plt.scatter(X_uw_uc, y_uw_uc, color = 'blue', alpha=0.5, )
plt.title('Unweighted data')

plt.scatter(X_w_uc, y_w_uc, color = 'red', alpha=0.5, )


import numpy as np

from phylokrr.kernels import *


class LKRR:

    def __init__(self, kernel = 'rbf') -> None:

        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        # internal
        self.alpha = []
        self.X = []
        self.y = []

    def p1(self, X, beta):
        return 1/(1 + np.exp(X.dot(beta)))

    def p0(self, X, beta):
        EX = np.exp(X.dot(beta))
        return EX/(1 + EX)

    def set_params(self, params):
        self.params = params

    def get_params(self):
        return self.params
    
    def fit(self, X, y):
        
        self.X = X
        self.y = y

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])

        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, self.y, self.params['lambda'])

    def predict(self, X_test):

        assert len(self.alpha) > 0, "The model needs to be fitted first"

        if self.kernel == 'rbf':
            K_test = RBF_kernel(X_test, self.X, self.params['gamma'])

        else:
            K_test = linear_kernel(X_test, self.X, self.params['c'])
        
        return K_test @ self.alpha
    
    def score(self, X_test, y_test, metric = 'rmse'):

        y_pred = self.predict(X_test)

        if metric == 'rmse':
            return np.sqrt( np.mean( (y_pred - y_test)**2 ) )
    
        else:
            # r2
            u = ((y_test - y_pred)**2).sum()
            v = ((y_test - y_test.mean())** 2).sum()

            return 1 - (u/v)
    
    def rmse(self, K, alpha, Y):

        return np.sqrt( np.mean( (K.dot(alpha) - Y)**2 ) )
    
    def opt_alpha(self, K, y, reg_lam = None):
        # Y = y
        # X = X
        # m = None
        K_Idx = K + reg_lam * np.diag( np.ones( K.shape[0] ) )
        return np.linalg.inv( K_Idx ).dot( y )



# set.seed(666)
# > x1 = rnorm(1000)           # some continuous variables 
# > x2 = rnorm(1000)
# > z = 1 + 2*x1 + 3*x2        # linear combination with a bias
# > pr = 1/(1+exp(-z))         # pass through an inv-logit function
# > y = rbinom(1000,1,pr)      # bernoulli response variable
# > 
# > #now feed it to glm:
# > df = data.frame(y=y,x1=x1,x2=x2)
# > glm( y~x1+x2,data=df,family="binomial")



def P_mat(vcv, chol = False, corr = False):
    
    if corr:
        Kr = np.diag(1/np.sqrt(np.diag(vcv)))
        vcv = Kr @ vcv @ Kr

    if chol:
        P = np.linalg.cholesky( np.linalg.inv( vcv ) )

    else:
        Oinv = np.linalg.inv( vcv )
        L,Q  = np.linalg.eig( Oinv )
        P  = Q @ np.diag( np.sqrt( 1/L ) ) @ Q.T

    return P


# ## logistic data

import glob
covs = glob.glob("/Users/ulises/Desktop/Machine Learning/HW3/SantaquiteriasProblem/simulations/rcoal_*.csv")

cov_file = covs[0]
vcv = np.loadtxt(cov_file, delimiter = ",")
nspps = vcv.shape[0]

mean_vector = np.zeros(nspps)

P = P_mat(vcv, chol= False, corr = False)

b = 1
X =  np.random.multivariate_normal(
    mean=mean_vector, 
    cov=vcv,
    # cov=np.diag(np.ones(n_spps)),
    size=2
    ).T

# logistic data
z = (X[:,0] + X[:,1])**b
probs = 1/(1 + np.exp(-z))
y = np.random.binomial(n = 1, p = probs)


sum(y == 1)
sum(y == 0)

np.max(probs)
np.min(probs)


import matplotlib.pyplot as plt
plt.scatter(X[:,0], X[:,1], c = y, alpha=0.6)


# y = rbinom(1000,1,pr) 


# data = np.loadtxt('../data/test_log_data.csv', delimiter=',')

# # get covariance matrix
# cov = np.loadtxt('../data/test_cov.csv',delimiter=',')


# [n,p] = np.shape(data)
# num_train = int(0.5*n)

# X,y = data[:,:-1], data[:,-1]
# P = P_mat(cov,corr=True)

# X = P @ X
# y = P @ y


num_train = round(nspps*0.60)

sample_train = X[0:num_train,:]
sample_test  = X[num_train: ,:]
    
label_train = y[0:num_train]
label_test  = y[num_train: ]
p = 2
    



import warnings

def p1(X, beta):
    return 1/(1 + np.exp(X.dot(beta)))

def p0(X, beta):
    EX = np.exp(X.dot(beta))
    return EX/(1 + EX)

def error_est(X, y,  beta, norm = False):
    if norm:
        return np.linalg.norm(p1(X,beta) - y, ord = 2)
        # pass
    else:
        y_pred = (p1(X, beta) >= 0.5).astype(float)
        return np.mean(y_pred != y)

def NewtonRaphsonStep(X, y, beta):

    Lp = -X.T @ ( y - p1(X, beta) )
    W  = np.diag( p1(X, beta) * p0(X, beta) )

    Lpp = -X.T @ W @ X
    try:
        return beta - np.linalg.inv(Lpp) @ Lp
    except np.linalg.LinAlgError:
        print(Lpp)
        return np.nan
    # return  beta - np.linalg.inv(Lpp) @ Lp


n_updts = 20
beta0 = np.array(np.random.normal(size = p ))
beta = np.copy(beta0)

epsilon = 1e-6
for i in range(n_updts):
    # i = 0
    beta_old = np.copy(beta)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        beta = NewtonRaphsonStep(sample_train, label_train, beta)

        if np.any( np.isnan(beta) ):
            print('beta re-redifined')
            beta = np.random.normal(size = p )

        error = error_est(sample_test, label_test, beta, norm=True)
        print("update = %i, error = %s" %(i, round(error, 5))) 

    if error <= epsilon:
        print('error convergence')
        break

    beta_norm = np.linalg.norm(beta_old - beta, ord = np.inf )
    if np.any( beta_norm <= epsilon  ):
        print('coefficient convergence')
        break

y_pred = (p1(sample_test, beta) >= 0.5).astype(float)

from sklearn.metrics import confusion_matrix
confusion_matrix(label_test,y_pred)




import random
import numpy as np

def split_data(X, y, num_test, seed = 123):
    random.seed(seed)
    n,_ = X.shape

    test_idx = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    return X_train, X_test, y_train, y_test

def split_data_vcv(X,y,vcv, num_test, seed = 123):
    
    random.seed(seed)
    n,_ = X.shape

    test_idx  = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]
    vcv_train, vcv_test = vcv[train_idx,:][:,train_idx], vcv[test_idx,:][:,test_idx]

    return X_train, X_test, y_train, y_test, vcv_train, vcv_test

def k_folds(X, folds = 3, seed = 123):
    """
    test_indx, train_indx
    """
    # X = X_train
    # folds = 4
    random.seed(seed)
    
    n,_ = X.shape
    all_index = list(range(n))
    random.shuffle(all_index)

    window = n/folds

    k_folds = []

    i = 0
    while True:

        init_indx = i
        end_indx  = i + window

        test_indx = all_index[round(init_indx):round(end_indx)]
        train_indx = list(set(all_index) - set(test_indx))
        # print(init_indx, end_indx)
        k_folds.append([test_indx, train_indx])

        i += window
        if i >= n:
            break

    # len(k_folds)
    return k_folds

def evaluate_folds(X, y, myFolds, model, tmp_params):

    # print(kwargs)
    # kwargs = {'c': 0.4, 'lambda': 0.1}
    # params = {'gamma': 0.4, 'lambda': 0.1}
    # params = tmp_params
    # model = phyloKRR(kernel='rbf')

    model.set_params(tmp_params)
    # model.get_params()

    all_errs = []
    for test_indx, train_indx in myFolds:
        # print(len(test_indx), len(train_indx))
        X_train,y_train = X[train_indx,:], y[train_indx]
        X_test,y_test = X[test_indx,:], y[test_indx]

        model.fit(X_train, y_train)

        # print(np.var(X_train))
        # print(np.var(model.X))
        # print(np.var(model.alpha))

        tmp_err = model.score(X_test, y_test, metric = 'rmse')
        all_errs.append(tmp_err)

    # return np.mean(all_errs)
    return np.median(all_errs)

def k_fold_cv_vcv(X, y, vcv, model, num_folds):
    """
    k-fold cross-validation with covariance matrix
    """
    n, p = X.shape
    fold_size = n // num_folds
    mse_sum = 0

    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]

        vcv_train = vcv[train_idx,:][:,train_idx]
        vcv_test = vcv[test_idx,:][:,test_idx]

        model.fit(X_train, y_train, vcv = vcv_train)
        mse_sum += model.score(X_test, y_test, vcv_test)

    return mse_sum / num_folds

def k_fold_cv_random_vcv(X, y, vcv,
                     model, 
                     params,
                     folds = 3, 
                     sample = 500,
                     verbose = True,
                     seed = 123
                     ):
    """
    Random search for hyperparameter tuning using k-fold cross-validation
    and covariance matrix
    """
    
    np.random.seed(seed=seed)
    # make random choice from the grid of hyperparameters
    all_params = params.keys()
    tested_params = np.zeros((sample, len(all_params)))
    for n,k in enumerate(all_params):
        tested_params[:,n] = np.random.choice(params[k], sample)

    if verbose:
        # check tested_params are unique
        tested_params = np.unique(tested_params, axis=0)
        print("Number of unique hyperparameters: ", tested_params.shape[0])
    
    all_errors = []
    for vec in tested_params:
        tmp_params = dict(zip(all_params, vec))
        model.set_params(**tmp_params)
        tmp_err = k_fold_cv_vcv(X, y, vcv, model, folds)
        all_errors.append([tmp_params, tmp_err])

    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]

    if verbose:
        print("CV score: ", best_[1])

    return best_[0]

def k_fold_cv(X, y, model, num_folds):
    """
    k-fold cross-validation
    """

    n, p = X.shape
    fold_size = n // num_folds

    all_errors = []
    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]
    
        model.fit(X_train, y_train)
        tmp_err = model.score(X_test, y_test, metric='rmse')
        all_errors.append(tmp_err)

    return np.mean(all_errors)

def k_fold_cv_random(X, y, model, params, folds=3, sample=500, verbose=True, seed=123):
    """
    Random search for hyperparameter tuning using k-fold cross-validation
    """
    np.random.seed(seed=seed)

    # make random choice from the grid of hyperparameters
    all_params = params.keys()
    tested_params = np.zeros((sample, len(all_params)))

    for n, k in enumerate(all_params):
        tested_params[:, n] = np.random.choice(params[k], sample)

    if verbose:
        # check tested_params are unique
        tested_params = np.unique(tested_params, axis=0)
        print("Number of unique hyperparameters: ", tested_params.shape[0])

    # shuffle the data
    n = X.shape[0]
    idx = np.arange(n)
    np.random.shuffle(idx)
    X = X[idx]
    y = y[idx]

    all_errors = []
    for vec in tested_params:
        tmp_params = dict(zip(all_params, vec))
        model.set_params(**tmp_params)
        tmp_err = k_fold_cv(X, y, model, folds)
        all_errors.append([tmp_params, tmp_err])

        if verbose:
            print("CV score: %s, Parameters %s" % (tmp_err, tmp_params))

    # take the best hyperparameters
    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]

    if verbose:
        print("Best CV score: ", best_[1])

    return best_[0]


def _scaler(ref, dat, use_sd = False):
    """
    Center and scale data
    """

    u = np.mean(ref, axis=0)
    centered = dat - u

    if use_sd:
        sd = np.std(ref, axis=0)
        sted = centered / sd

        return sted
    else:

        return centered

def weight_data(X_uw_uc, y_uw_uc, vcv, use_sd = False):
    """
    Weight data using the square root of the inverse of the covariance matrix
    and scale the data

    Parameters
    ----------
    X_uw_uc: np.array
        unweighted independent variables (uncentered)
    
    y_uw_uc: np.array
        unweighted response variable (uncentered)

    vcv: np.array
        covariance matrix

    use_sd: bool
        whether to scale the data using the standard deviation

    Returns
    -------
    X_w: np.array
        weighted and scaled independent variables

    y_w: np.array
        weighted and scaled response variable
    """

    P = P_mat_simple(vcv)

    X_w_uc = P @ X_uw_uc
    y_w_uc = P @ y_uw_uc

    X_w = _scaler(X_w_uc, X_w_uc, use_sd)
    y_w = _scaler(y_w_uc, y_w_uc, use_sd)

    return X_w, y_w

class PGLS:
    def __init__(self, 
                 fit_intercept=False) -> None:
        
        self.fit_intercept = fit_intercept
        # self.weighted = weighted
        self.intercept = 0
        self.beta = np.array([])

    def std_PGLS(self, X, y, vcv):
        """
        Generalized Least Squares with phylogenetic covariance matrix
        as the weight matrix
        """
        n,p = X.shape
        # Oinv = np.linalg.inv(vcv)

        if self.fit_intercept:
            X = np.hstack((np.ones((n,1)), X))

        self.beta =  np.linalg.solve(
                        X.T @ np.linalg.solve( vcv , X ),
                        X.T @ np.linalg.solve( vcv , y )
                    )

    def OLS(self, X, y):
        """
        Ordinary Least Squares
        """
        n,p = X.shape

        if self.fit_intercept:
            X = np.hstack((np.ones((n,1)), X))

        self.beta = np.linalg.solve(X.T @ X, X.T @ y)

    def fit(self, X, y, vcv=None):
        """
        if vcv is not None, then we use phylogenetic covariance matrix
        for Generalized Least Squares. Otherwise, we use Ordinary Least Squares
        """

        if vcv is not None:
            self.std_PGLS(X, y, vcv)

        else:
            self.OLS(X, y)

    def predict(self, X):

        n, p = X.shape
        if self.fit_intercept:
            X = np.hstack((np.ones((n, 1)), X))

        return X @ self.beta

    def score(self, X_test, y_test, metric='rmse'):
        y_pred = self.predict(X_test)

        if metric == 'rmse':
            return np.sqrt(np.mean((y_pred - y_test) ** 2))
        else:
            u = ((y_test - y_pred) ** 2).sum()
            v = ((y_test - y_test.mean()) ** 2).sum()

            return 1 - (u / v)

def P_inv_simple(vcv):
    """
    get the square root of the inverse of the
    """

    # Kr = np.diag(1/np.sqrt(np.diag(vcv)))
    # vcv = Kr @ vcv @ Kr

    if not isinstance(vcv, np.ndarray):
        return None

    L,Q = np.linalg.eig( vcv )
    P = Q @ np.diag( L**(1/2) ) @ Q.T

    return P

def P_mat_simple(vcv):
    """
    get the square root of the inverse of the
    """

    # Kr = np.diag(1/np.sqrt(np.diag(vcv)))
    # vcv = Kr @ vcv @ Kr

    if not isinstance(vcv, np.ndarray):
        return None

    L,Q = np.linalg.eig( vcv )
    P = Q @ np.diag( L**(-1/2) ) @ Q.T

    return P

def myfunc(x, b=0, type='sin', add_noise = False, noise = 1):    

    if type == 'sin':
        y =  np.sin( np.sum(x, axis=1) * b).ravel()

    else:
        y =  np.sum(x, axis=1) ** b

    if add_noise:
        y += noise

    return y


def get_Xy_w_uc(X_uw_uc, P, b, type, on_weighted, add_noise, noise):    

    if on_weighted:

        X_w_uc = P @ X_uw_uc
        y_w_uc = myfunc(X_w_uc, b=b, type=type).ravel()

        if add_noise:
            y_w_uc += noise

    else:
        y_uw_uc = myfunc(X_uw_uc, b=b, type=type).ravel()

        if add_noise:
            y_uw_uc += noise        

        X_w_uc = P @ X_uw_uc
        y_w_uc = P @ y_uw_uc

    return X_w_uc, y_w_uc

def get_Xy_wc(X_uw_uc, P_mat, b, type, on_weighted, add_noise, noise):    

    if on_weighted:
        
        X_w_uc = P_mat @ X_uw_uc
        y_w_uc = myfunc(X_w_uc, b=b, type=type).ravel()

        if add_noise:
            y_w_uc += noise

    else:
        y_uw_uc = myfunc(X_uw_uc, b=b, type=type).ravel()

        if add_noise:
            y_uw_uc += noise        

        X_w_uc = P_mat @ X_uw_uc
        y_w_uc = P_mat @ y_uw_uc

    X_wc = _scaler(X_w_uc, X_w_uc, use_sd= not False)
    y_wc = _scaler(y_w_uc, y_w_uc, use_sd=  False)

    return X_wc, y_wc

def sim_data_signal(vcv, mean_vector, b, 
             on_weighted=False, 
             add_noise = False, 
             noise_var = 1,
             n_var = 3, 
             extra_weights = [1,1,1], 
             type='pol'):
    """
    Simulate data for n_var independent variables

    Parameters
    ----------
    vcv: np.array
        covariance matrix where data is simulated from
    
    mean_vector: np.array
        mean vector
    
    b: int
        power of the polynomial function

    on_weighted: bool
        whether to use weighted data or not to obtain the response variable.
        the weights are obtained using the 
        squared root of the inverse of 
        the covariance matrix

    add_noise: bool
        whether to add noise to the response variable

    noise_var: float
        variance of the noise

    extra_weights: list
        weights for the independent variables

    type: str
        type of the function to simulate the data
        if 'sin' then a sine function is used over the
        sum of the independent variables
        if 'pol' then a polynomial function is used over the
        sum of the independent variables
    
    weight_with_vcv: bool
        if true (default) the weighting of the values is 
        done using vcv. Otherwise, vcv2 is used

    vcv2: np.array
        if weight_with_vcv is false, then we use this covariance matrix
        to weight the data
    
    Returns
    -------
    X_w_uc: np.array
        weighted independent variables (uncentered)
    
    y_w_uc: np.array
        weighted response variable (uncentered)
    """
    # n_var = 3

    assert len(extra_weights) == n_var, "extra weights must have the same length as the number of predictors"

    X_uw_uc = np.zeros((vcv.shape[0], n_var))
    for j in range(n_var):
        X_uw_uc[:,j] = np.random.multivariate_normal(mean=mean_vector, cov=vcv)*extra_weights[j]


    noise = np.random.normal(0, noise_var, X_uw_uc.shape[0])


    P1 = P_mat_simple(vcv)
    P2 = P_mat_simple(np.diag(np.diag(vcv)))
    
    (X_w_uc_1, y_w_uc_1) = get_Xy_w_uc(X_uw_uc, P1, b, type,
                                       on_weighted,
                                       add_noise, noise)

    (X_w_uc_2, y_w_uc_2) = get_Xy_w_uc(X_uw_uc, P2, b, type,
                                       on_weighted,
                                       add_noise, noise)

    return (X_w_uc_1, X_w_uc_2, y_w_uc_1, y_w_uc_2)




import numpy as np
from phylokrr.core import PhylogeneticRegressor


def dim_red(X, n_com):
    # extracting averages
    X_centered = (X - np.mean(X, axis=0))

    # U,E,Vt (SVD)
    _,_,Vt = np.linalg.svd(X_centered)
    # PCA components
    W = Vt.T[:, :n_com]
    # projection into n_com dimensions
    return X_centered.dot(W)



# get data
data = np.loadtxt('../data/test_data.csv', delimiter=',')
X,y = data[:,:-1], data[:,-1]


X1D = dim_red(X, 1)

import matplotlib.pyplot as plt
plt.scatter(X1D[:,0], y,)


# get covariance matrix
cov = np.loadtxt('../data/test_cov.csv',delimiter=',')

# set model
model = PhylogeneticRegressor(X, y, cov, kernel='rbf')

# set hyperparamter space
params = {
    'lambda' : np.logspace(-5, 15, 200, dtype = float, base=10),
    'gamma' : np.logspace(-15, 3, 200, dtype = float, base=10),
}
model.set_hyperparameter_space(params=params)
    
# fit the model
model.fit(seed=12, sample=200)
model


# also you can set new set of hyperparameters
# and then fit
model.set_params(
    {'lambda': 0.0001, 'gamma': 0.0755},
    fit=True
)

# plotting
import matplotlib.pyplot as plt


feature_names = ['feature 1', 'feature 2', 'feature 3']

FI = model.FeatureImportance(seed=None)
plt.bar(feature_names, FI)
plt.ylabel('RMSE increase')

# len(model.P_inv)
# partial depedence plot

pdp_0 = model.pdp(feature=0)
pdp_1 = model.pdp(feature=1)
pdp_2 = model.pdp(feature=2)



plt.rcParams['figure.figsize'] = [10, 4]

fig, (ax1, ax2, ax3) = plt.subplots(1, 3)
fig.suptitle('Partial Dependence Plot')

ax1.scatter(pdp_0[:,0], pdp_0[:,1])
ax1.set_ylabel('Predicted y')
ax1.set_xlabel('feature 1')

ax2.scatter(pdp_1[:,0], pdp_1[:,1])
ax2.set_xlabel('feature 2')

ax3.scatter(pdp_2[:,0], pdp_2[:,1])
ax3.set_xlabel('feature 3')



f1_idx = 2
f2_idx = 1

main_data = model.X

f1 = model.sample_feature(f1_idx, quantiles=[0,1], sample=30)
f2 = model.sample_feature(f2_idx, quantiles=[0,1], sample=30)


# make a meshgrid from f1 and f2
X1, X2 = np.meshgrid(f1, f2)
OUT = np.zeros(X1.shape)

for i in range(X1.shape[0]):
    for j in range(X1.shape[1]):
        # i,j
        X_tmp = main_data.copy()
        X_tmp[:,f1_idx] = X1[i,j]
        X_tmp[:,f2_idx] = X2[i,j]
        OUT[i,j] = np.mean( model.P_inv @ model.model.predict(X_tmp))

# can you add contour lines?

plt.rcParams['figure.figsize'] = [5, 5]
plt.contourf(X1, X2, OUT)
plt.colorbar()
plt.xlabel('feature 1')
plt.ylabel('feature 2')


plt.title('2D Partial Dependence Plot')



# ## 
# data = np.loadtxt('../data/test_log_data.csv', delimiter=',')
# X,y = data[:,:-1], data[:,-1]
# import matplotlib.pyplot as plt
# plt.scatter(X[:,0], X[:,2], c = y, alpha=0.6)


# file_path = "/Users/ulises/Desktop/ABL/software/phylokrr/data/test.tree"

# import dendropy
# tree = dendropy.Tree.get( path = 'path/to/tree', schema = 'newick')

# for i in tree.postorder_node_iter():
#     print( i.distance_from_root() )






import random
import numpy as np

# seed for reproducibility
np.random.seed(12038)
random.seed(12038)

from phylokrr.kernels import RBF_kernel, linear_kernel
from phylokrr.utils   import PGLS
from phylokrr.kernels import R2,rmse

def P_mat_simple(vcv):
    """
    get the square root of the inverse of the
    covariance matrix using spectral decomposition

    vcv: np.array, default = None
        covariance matrix
    """
    
    if isinstance(vcv, type(None)):
        return None

    L,Q  = np.linalg.eig( vcv )
    P  = Q @ np.diag( L**(-1/2) ) @ Q.T 

    return P

def _scaler(ref, dat, use_sd = False):
    """
    Center and scale data

    ref: np.array
        reference data
    
    dat: np.array
        data to be scaled
    
    use_sd: bool, default = False
        use standard deviation to scale data
    
    Returns:
    --------
    np.array
        scaled data
    """

    u = np.mean(ref, axis=0)
    centered = dat - u

    if use_sd:
        sd = np.std(ref, axis=0)
        sted = centered / sd
        return sted
    
    else:
        return centered

def scale_weighted_data(X, vcv, use_sd = False):
    """
    scale data by the square root of the inverse of the
    covariance matrix

    X: np.array
        data

    vcv: np.array
        covariance matrix
    """
    P = P_mat_simple(vcv)
    X_w_uc =  P @ X
    return _scaler(X_w_uc, X_w_uc, use_sd)


class PGLS:
    def __init__(self, fit_intercept = False) -> None:
        self.fit_intercept = fit_intercept
        self.beta = np.array([])
    
    def fit(self, X, y, vcv):
        n,p = X.shape
        Oinv = np.linalg.inv(vcv)
        if self.fit_intercept:
            X_ = np.hstack((np.ones((n,1)), X))
        else:
            X_ = X

        self.beta = np.linalg.inv(X_.T @ Oinv @ X_) @ X_.T @ Oinv @ y
    
    def predict(self, X):
        if self.fit_intercept:
            X_ = np.hstack((np.ones((X.shape[0],1)), X))
        else:
            X_ = X
        return X_ @ self.beta

class WKRR:

    def __init__(self, kernel = 'rbf', fit_intercept = True, check_cov = False) -> None:
        """
        Weighted Kernel Ridge Regression

        kernel: str, default = 'rbf'
            kernel type

        fit_intercept: bool, default = True
            fit intercept term

        check_cov: bool, default = False
            Check covariance matrix, i.e., P.T @ P == vcv^-1
        """

        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        self.fit_intercept = fit_intercept
        self.check_cov = check_cov
        
        # internal
        self.intercept = 0
        self.alpha = np.array([])
        self.X = np.array([])

        self.chol = False

    def set_params(self, **params):

        if self.kernel == 'rbf':
            self.params['gamma'] = params['gamma']

        else:
            self.params['c'] = params['c']

        self.params['lambda'] = params['lambda']

    def get_params(self):
        return self.params
    
    def P_mat(self, vcv):
        """
        get the square root of the inverse of the
        covariance matrix.

        vcv: np.array, default = None
            covariance matrix
        """
        
        if isinstance(vcv, type(None)):
            return None
        
        if self.check_cov:
            self.assert_COV_sym(vcv)

        if self.chol:
            C = np.linalg.cholesky( vcv )
            P = np.linalg.inv( C )

        else:
            L,Q  = np.linalg.eig( vcv )
            P  = Q @ np.diag( L**(-1/2) ) @ Q.T 

        if self.check_cov:
            self.assert_COV_decom(P, vcv)

        return P
    
    def assert_COV_sym(self, vcv, tol=1e-8):
        assert np.all(np.abs(vcv-vcv.T) < tol), 'not symmetric matrix'

    def assert_COV_decom(self, P, vcv):
        """
        check if P.T @ P == vcv^-1
        expensive calculation for large matrices
        """
        assert np.all(np.round(P.T @ P, 2) == np.round(np.linalg.inv(vcv), 2)), "P.T @ P != vcv^-1"

    def fit(self, X, y, vcv = None):
        """
        Fit the model

        vcv: np.array, default = None
            covariance matrix

        """
        # X = X_train
        # y = y_train
        # vcv = vcv_train
        
        self.X = X
        # self.y = y
        P = self.P_mat(vcv)

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])

        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, y, self.params['lambda'], P)
        
        # intercept term 
        if self.fit_intercept:
            self.intercept = np.mean(y - K_train @ self.alpha)

    def predict(self, X_test):

        assert len(self.alpha) > 0, "The model needs to be fitted first"

        if self.kernel == 'rbf':
            K_test = RBF_kernel(X_test, self.X, self.params['gamma'])

        else:
            K_test = linear_kernel(X_test, self.X, self.params['c'])
        
        return K_test @ self.alpha + self.intercept
    
    def score(self, X_test, y_test, vcv_test, metric = 'rmse'):
        """
        
        Calculates score considerint the covariance matrix
        This error is the same as the in objective function

        X_test: np.array
            testing data
        
        y_test: np.array
            testing targets
        
        vcv_test: np.array
            testing covariance matrix. If None, then it
            will return the standard RMSE and R^2

        metric: str, default = 'rmse'
            weighted root mean squared error (the same 
            that was used to construct the objective function).
            if metric = 'r2', then it returns the R^2 which
            also considers the covariance matrix, see Willett & Singer (1988)

        Returns:
        --------
        float
            score

        Reference:
        ---------
        Willett, J. B., & Singer, J. D. (1988). 
        Another cautionary note about R 2: Its use 
        in weighted least-squares regression analysis. 
        The American Statistician, 42(3), 236-238.
        """

        y_pred = self.predict(X_test)
        r = y_pred - y_test
        
        if isinstance(vcv_test, type(None)):
            P = np.eye(X_test.shape[0])

        else:
            P = self.P_mat(vcv_test)

        werr = P @ r # weighted residuals
        Py   = P @ y_test # weighted targets

        if metric == 'rmse':
            return rmse(werr)
        
        else:
            return R2(Py, werr)
            
        # if metric == 'rmse':
        #     return np.sqrt( np.mean( r**2 ) )
        
        # else:
        #     u = r.T @ r
        #     v = y_test.T @ y_test - n * np.mean(y_test)**2
        #     return 1 - (u/v)
    
    def opt_alpha(self, K, y, reg_lam = None, P = None):
        # Y = y
        # X = X
        # m = None
        n,_ = self.X.shape
        
        I = np.eye(K.shape[0])
        nlI = n * reg_lam * I

        if isinstance(P, type(None)):
            return np.linalg.solve(K + nlI, y)
        
        else:
            return P @ np.linalg.solve(P @ K @ P + nlI,  P @ y)
            # return np.linalg.solve(P.T @ P @ K + nlI,  P.T @ P @ y)

class KRR(WKRR):
    
        def __init__(self, kernel = 'rbf', fit_intercept = True) -> None:
            """
            Kernel Ridge Regression
    
            kernel: str, default = 'rbf'
                kernel type
    
            fit_intercept: bool, default = True
                fit intercept term
            """
    
            super().__init__(kernel, fit_intercept)
                    # internal
            self.intercept = 0
            self.alpha = np.array([])
            self.X = np.array([])


        def fit(self, X, y):
        
            self.X = X
            self.y = y

            if self.kernel == 'rbf':
                K_train = RBF_kernel(self.X, self.X, self.params['gamma'])

            else:
                K_train = linear_kernel(self.X, self.X, self.params['c'])

            self.alpha = self.opt_alpha(K_train, self.y, self.params['lambda'])

            # intercept term 
            if self.fit_intercept:
                self.intercept = np.mean(y - K_train @ self.alpha)            

        def score(self, X_test, y_test, metric = 'rmse'):

            y_pred = self.predict(X_test)

            if metric == 'rmse':
                return np.sqrt( np.mean( (y_pred - y_test)**2 ) )
        
            else:
                # r2
                u = ((y_test - y_pred)**2).sum()
                v = ((y_test - y_test.mean())** 2).sum()

                return 1 - (u/v)        
        
        def opt_alpha(self, K, y, reg_lam = None):
            # Y = y
            # X = X
            # m = None
            n,_ = self.X.shape
            K_Idx = K + n * reg_lam * np.diag( np.ones( K.shape[0] ) )
            return np.linalg.inv( K_Idx ).dot( y )


# cov. matrix obtained from the phylogenetic tree
vcv = np.loadtxt("../data/test_cov2.csv", delimiter=',') 

# Trait simulation under Brownian motion
n = vcv.shape[0]
mean = np.zeros(n)
X = np.random.multivariate_normal(cov=vcv, mean=mean).reshape(-1,1)
# Non-linear response variable (sine curve)

X = scale_weighted_data(X, vcv, use_sd = True)
y = np.sin(X*2).ravel() + 5 

# Add noise to the response variable
y[::10] += 4 * (0.5 - np.random.rand(X.shape[0] // 10)) 
# y = scale_weighted_data(y.reshape(-1,1), vcv, use_sd = False)

num_test = round(0.5*n)
from phylokrr.utils import split_data

(X_train, y_train, X_test, y_test) = split_data(X, y, num_test = num_test, seed = 12038)

model = KRR(kernel='rbf', fit_intercept= True)



# arbitrarily proposed hyperparameters
params = {'lambda': 1, 'gamma': 1}

# set hyperparamters
model.set_params(**params)


# fit model with phylogenetic covariance matrix
model.fit(X_train, y_train)
y_pred1 = model.predict(X_test)
rmse1 = model.score(X_test, y_test,  metric='rmse')
r1 = model.score(X_test, y_test,  metric='r2')







import matplotlib.pyplot as plt
# fit standard phylogenetic regression
b_wls = PGLS(X_train, y_train, vcv_train)
y_pred3 = np.hstack((np.ones((X_test.shape[0],1)), X_test)) @ b_wls
P = model.P_mat(vcv_test)
w_r_pgls = P @ (y_pred3 - y_test)
Py_test = P @ y_test
r2_pgls = R2(Py_test, w_r_pgls)
rmse_pgls = rmse(w_r_pgls)
# store figure

plt.scatter(X_test, y_test , color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
plt.scatter(X_test, y_pred1, color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
# plt.scatter(X_test, y_pred2, color = 'gray', alpha=0.5, label = 'KRR predictions w\o CV')
plt.scatter(X_test, y_pred3, color = 'red', alpha=0.5, label = 'PGLS predictions')
plt.xlabel('x')
plt.ylabel('y')
plt.legend()
plt.tight_layout()








from phylokrr.utils import split_data_vcv

# split data into training and testing sets 
num_test = round(0.5*n)

(X_train  , X_test,  
 y_train  , y_test,  
 vcv_train, vcv_test) = split_data_vcv(X, y, vcv, num_test, seed = 12038) # seed defined above

from phylokrr.kernels import KRR

# set model
model = KRR(kernel='rbf', fit_intercept= True)

# arbitrarily proposed hyperparameters
params = {'lambda': 1, 'gamma': 1}

# set hyperparamters
model.set_params(**params)

# fit model with phylogenetic covariance matrix
model.fit(X_train, y_train, vcv = vcv_train)
y_pred1 = model.predict(X_test)
rmse1 = model.score(X_test, y_test, vcv_test, metric='rmse')
r1 = model.score(X_test, y_test, vcv_test, metric='r2')

model.fit(X_train, y_train, vcv =None)
y_pred2 = model.predict(X_test)
rmse2 = model.score(X_test, y_test, None, metric='rmse')
r2 = model.score(X_test, y_test, None, metric='r2')


import matplotlib.pyplot as plt


# fit standard phylogenetic regression
b_wls = PGLS(X_train, y_train, vcv_train)
y_pred3 = np.hstack((np.ones((X_test.shape[0],1)), X_test)) @ b_wls
P = model.P_mat(vcv_test)
w_r_pgls = P @ (y_pred3 - y_test)
Py_test = P @ y_test
r2_pgls = R2(Py_test, w_r_pgls)
rmse_pgls = rmse(w_r_pgls)
# store figure

plt.scatter(X_test, y_test , color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
plt.scatter(X_test, y_pred1, color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
# plt.scatter(X_test, y_pred2, color = 'gray', alpha=0.5, label = 'KRR predictions w\o CV')
plt.scatter(X_test, y_pred3, color = 'red', alpha=0.5, label = 'PGLS predictions')
plt.xlabel('x')
plt.ylabel('y')
plt.legend()
plt.tight_layout()
plt.savefig(
    './../data/imgs/phyloKRR_vs_PGLS.png', dpi = 200,
    bbox_inches = 'tight', pad_inches = 0.1
    )



from phylokrr.utils import k_fold_cv_random

params = {
    'lambda' : np.logspace(-3, 3, 200, base=2),
    'gamma' : np.logspace(-3, 3, 200,  base=2),
}

best_params = k_fold_cv_random(X_train, y_train, vcv_train,
                                model, 
                                params,
                                folds = 3, 
                                sample = 70)

model.set_params(**best_params)
model.fit(X_train, y_train, vcv = vcv_train)
y_pred_cv = model.predict(X_test)


# control the size
fs = 18
plt.rcParams['figure.figsize'] = [7, 6]
plt.scatter(X_test, y_test, color = 'blue' ,
            alpha=0.5, 
            label = 'Testing (unseen) data',)
plt.scatter(X_test, y_pred_cv, 
            color = 'green', alpha=0.5, 
            label = 'phyloKRR predictions',)
plt.scatter(X_test, y_pred3, 
            color = 'red', alpha=0.5, 
            label = 'PGLS predictions',)
plt.xlabel('x', fontsize = fs)
plt.ylabel('y', fontsize = fs)
plt.legend(fontsize = fs)
plt.tight_layout()
plt.savefig(
    '/Users/ulises/Desktop/ABL/phyloKRR_vs_PGLS_cv.png', dpi = 200,
    bbox_inches = 'tight', pad_inches = 0.1
    )


r2_cv = model.score(X_test, y_test, vcv_test, metric='r2')

model.score(X_test, y_test, vcv_test, metric='rmse')



import numpy as np
import pandas as pd
from scipy.stats import multivariate_normal
from scipy.special import expit  # Sigmoid function
import statsmodels.api as sm
from sklearn.metrics import brier_score_loss
from sklearn.calibration import calibration_curve
import matplotlib.pyplot as plt


# -log P(yt|yp) = -(yt log(yp) + (1 - yt) log(1 - yp))

# Parameters
n_obs = 100   # Number of observations (within one cluster)
beta = np.array([1, 2, 3])  # Coefficients (Intercept, x1, x2)
rho = 0.5  # Correlation between responses

# Generate Covariates
np.random.seed(12038)
x1 = np.random.normal(size=n_obs)  # Continuous predictor
x2 = np.random.binomial(1, 0.5, size=n_obs)  # Binary predictor
X = np.column_stack((np.ones_like(x1), x1, x2))  # Design matrix with intercept

# Generate Correlated Errors
cov_matrix = rho * np.ones((n_obs, n_obs)) + (1 - rho) * np.eye(n_obs)  # Exchangeable correlation
correlated_errors = multivariate_normal.rvs(mean=np.zeros(n_obs), cov=cov_matrix)

# Compute Probabilities
eta = X @ beta  # Linear predictor
p = expit(eta)  # Convert to probability

# Generate Correlated Binary Outcomes using Thresholding
y_null = (p + correlated_errors > np.median(p + correlated_errors)).astype(int)

# Create DataFrame
df = pd.DataFrame({"x1": x1, "x2": x2, "y": y_null})

# Fit Models
logit_result_null = sm.Logit(y_null, sm.add_constant(df[["x1", "x2"]])).fit()
gee_result_null = sm.GEE(y_null, sm.add_constant(df[["x1", "x2"]]), groups=np.ones(len(df)), 
                         family=sm.families.Binomial(), cov_struct=sm.cov_struct.Exchangeable()).fit()

# Extract p-values
logit_pvalues = logit_result_null.pvalues
gee_pvalues = gee_result_null.pvalues

# Compare Results
print("Simple Logistic Regression Coefficients & Standard Errors:")
print(logit_result_null.summary())

print("\nGEE Coefficients & Robust Standard Errors:")
print(gee_result_null.summary())

# Calculate Brier Score
logit_probs = logit_result_null.predict()
gee_probs = gee_result_null.predict()
brier_score_logit = brier_score_loss(y_null, logit_probs)
brier_score_gee = brier_score_loss(y_null, gee_probs)
print("\nBrier Score for Logistic Regression:", brier_score_logit)
print("Brier Score for GEE:", brier_score_gee)

# Plot Calibration Curves
plt.figure(figsize=(7, 5))
prob_true_logit, prob_pred_logit = calibration_curve(df["y"], logit_probs, n_bins=10)
prob_true_gee, prob_pred_gee = calibration_curve(df["y"], gee_probs, n_bins=10)

plt.plot(prob_pred_logit, prob_true_logit, "s-", label="Logistic Regression")
plt.plot(prob_pred_gee, prob_true_gee, "o-", label="GEE")
plt.plot([0, 1], [0, 1], "k--")  # Perfect calibration
plt.xlabel("Predicted Probability")
plt.ylabel("True Probability")
plt.title("Calibration Curve")
plt.legend()
plt.show()



import numpy as np
import matplotlib.pyplot as plt

from phylokrr.utils import k_fold_cv_random, split_data, PGLS
from phylokrr.datasets import load_vcv, sim_data
from phylokrr.kernels import KRR
np.random.seed(12038)




vcv = load_vcv()
n = vcv.shape[0]
mean_vector = np.zeros(n)
num_test = round(0.40 * n)

# simulate phylogenetically weighted data
X_w, y_w = sim_data(vcv, mean_vector, b=1, n_var=1)


(X_train, X_test,
 y_train, y_test,) = split_data(X_w, y_w, num_test, seed = 12038) # seed defined above



rbf_grid_params = {
    'lambda': np.logspace(-15, 15, 200, dtype=float, base=2),
    'gamma' : np.logspace(-15, 15, 200, dtype=float, base=2),
}

krr_rbf = KRR(kernel='rbf', fit_intercept=True)
pgls_model = PGLS(fit_intercept =  True)

best_params = k_fold_cv_random(X_train, y_train,
                                krr_rbf,
                                rbf_grid_params,
                                verbose = False,
                                folds = 5,
                                sample = 100)

krr_rbf.set_params(**best_params)
krr_rbf.fit(X_train, y_train)
y_pred_cv = krr_rbf.predict(X_test)

pgls_model.fit(X_train, y_train)
y_pred_pgls = pgls_model.predict(X_test)



fs = 10
fig, axs = plt.subplots(1, 2, figsize=(8, 4))
axs[0].scatter(X_test, y_test, color = 'blue', 
               alpha=0.5, label = 'Testing (unseen) data',)
axs[0].scatter(X_test, y_pred_pgls, color = 'red', 
               alpha=0.5, label = 'PGLS predictions',)
axs[0].legend()
axs[1].scatter(X_test, y_test, color = 'blue', alpha=0.5, 
               label = 'Testing (unseen) data',)
axs[1].scatter(X_test, y_pred_cv, color = 'green', alpha=0.5, 
               label = 'RBF kernel predictions',)
axs[1].legend()


r2_kernel = krr_rbf.score( X_test, y_test, metric='r2')
r2_pgls = pgls_model.score(X_test, y_test, metric='r2')

print(f"R2 for phyloKRR: {r2_kernel}")
print(f"R2 for PGLS: {r2_pgls}")



import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import cholesky


def pre_process_V(V):
    n, _ = V.shape
    V = V/np.max(V)
    detV_n = np.linalg.det(V)**(1/n)
    print('detV_n: ', detV_n)
    V = V/detV_n if detV_n != 0 else V
    return V

def make_circles(rho, num_samples, seed, test_num = 0.5, sigma=1, radius=0.01):

    np.random.seed(seed)
    # # Step 1: Define correlation structure for observations
    # num_samples = 5
    # rho = 0.5  # Strength of correlation between consecutive observations

    # Construct correlation matrix (exponential decay)
    V = np.array([[rho**abs(i-j) for j in range(num_samples)] for i in range(num_samples)])

    # Step 2: Generate independent Gaussian features
    # noise = 0.1  # Add noise to the observations
    num_features = 2
    X_uncorrelated = np.random.randn(num_samples, num_features) * sigma

    # Step 3: Apply Cholesky decomposition to correlate observations
    L = cholesky(V)
    X = L @ X_uncorrelated   # Now observations are correlated

    # Step 4: Define a **circular non-linearity** (donut shape)
    # radius sets the radius for the decision boundary

    decision_boundary = np.sqrt(X[:, 0]**2 + X[:, 1]**2)  # Compute distance from origin
    y = ((decision_boundary > radius) & (decision_boundary < radius + 1)).astype(int)  # Define labels

    # Step 5: Shuffle the data and split into training and testing sets
    all_index = list(range(X.shape[0]))
    np.random.shuffle(all_index)
    
    X = X[all_index]
    y = y[all_index]
    V = V[all_index, :][:, all_index]


    X = X[all_index,:]
    y = y[all_index]
    V = V[all_index, :][:, all_index]

    Indx = int(test_num*X.shape[0])
    test_idx = np.random.choice(all_index, Indx, replace=False)
    train_idx = list(set(all_index) - set(test_idx))

    X_train, X_test = X[train_idx, :], X[test_idx, :]
    y_train, y_test = y[train_idx], y[test_idx]

    V_train = V[train_idx,:][:,train_idx]
    V_test = V[test_idx,:][:,test_idx]
    
    return X_train, X_test, y_train, y_test, V_train, V_test


def plot_dataset(X, y, axes):
    plt.plot(X[:, 0][y==0], X[:, 1][y==0], "bs")
    plt.plot(X[:, 0][y==1], X[:, 1][y==1], "g^")
    plt.axis(axes)
    plt.grid(True, which='major', linestyle='--', linewidth='0.5')
    plt.xlabel(r"$x_1$", fontsize=20)
    plt.ylabel(r"$x_2$", fontsize=20, rotation=0)

def plot_predictions(clf, axes):
    x0s = np.linspace(axes[0], axes[1], 100)
    x1s = np.linspace(axes[2], axes[3], 100)
    x0, x1 = np.meshgrid(x0s, x1s)
    X = np.c_[x0.ravel(), x1.ravel()]
    y_pred = clf.predict(X).reshape(x0.shape)
    plt.contourf(x0, x1, y_pred, cmap=plt.cm.bwr , alpha=0.3)

from krgee import Logistic
import matplotlib.pyplot as plt
from cross_validation_krgee import (k_fold_cv_random_vcv_v2)

# seed = 1203
seed = 12038
cv_sample = 100
cv_folds = 4
tol = 1e-3
max_iter = 70

(X_train, X_test, 
 y_train, y_test, 
 vcv_train, vcv_test) = make_circles(
                                rho=0.5,
                                num_samples=500,
                                 seed=seed, 
                                 test_num=0.35, 
                                 sigma=1,
                                 radius=.1
                                 )


# plot_dataset(X_train, y_train, np.array([-1, 1, -1, 1])*4)
# plot_dataset(X_test, y_test, np.array([-1, 1, -1, 1])*4)
# vcv_train = pre_process_V(vcv_train)

print('X_train shape: ', X_train.shape)
print('X_test shape: ', X_test.shape)

print('label_ratio train:', np.sum(y_train ==1)/len(y_train))
print('label_ratio test:', np.sum(y_test ==1)/len(y_test))


X_train_offset = X_train.mean(axis=0)
X_train_scale = X_train.std(axis=0)
X_train -= X_train_offset
X_train /= X_train_scale

# X_test -= np.mean(X_test, axis=0)
# X_test /= np.std(X_test, axis=0)

X_test -= X_train_offset
X_test /= X_train_scale


params = {
    'lambda' : np.logspace(-7, 3, 200, base=2),
    'gamma'  : np.logspace(-7, 3, 200, base=2),
}


I_train = np.eye(X_train.shape[0])
lkrr = Logistic(get_corr=False, tol=tol, 
               max_iter=max_iter, kernel='rbf', 
               verbose=True, seed=seed, 
               fix_phi=True)
# lkrr.phi = 1 


import time
start = time.time()
lkrr_best_params = k_fold_cv_random_vcv_v2(
    X_train, y_train, I_train, 
    lkrr, params,
    folds=cv_folds,
    sample=cv_sample, 
    verbose=True, 
    seed=seed
)
print('lkrr CV Time: ', time.time() - start)


print('lkrr best params: ', lkrr_best_params)
lkrr.set_params(**lkrr_best_params)
lkrr.fit(X_train, y_train, I_train)
lkrr_loss = lkrr.score(X_test, y_test)
print('lkrr loss: ', lkrr_loss)
print('phi: ', lkrr.phi)
print('degrees of freedom: ', lkrr.dof)
print()


lkrgee = Logistic(get_corr=False, tol=tol, 
               max_iter=max_iter, kernel='rbf', 
               verbose=True, seed=seed, 
               fix_phi=True)



import time
start = time.time()
lkrgee_best_params = k_fold_cv_random_vcv_v2(
    X_train, y_train, vcv_train, 
    lkrgee, params,
    folds=cv_folds,
    sample=cv_sample, 
    verbose=True, 
    seed=seed
)
print('lkrgee CV Time: ', time.time() - start)

print('lkrgee best params: ', lkrgee_best_params)
lkrgee.set_params(**lkrgee_best_params)
# lkrgee.dof = emp_df
# lkrgee.verbose = True
# lkrgee._verbose = False
lkrgee.fit(X_train, y_train, vcv_train)
# print("R inv max: ", np.max(lkrgee.R_inv))
# print("R inv min: ", np.min(lkrgee.R_inv))
lkrgee_loss = lkrgee.score(X_test, y_test)
print('lkrgee_loss testing: ', lkrgee_loss)
print('phi: ', lkrgee.phi)
print('degrees of freedom: ', lkrgee.dof)
# print('rho: ', lkrgee.rho)





from sklearn.linear_model import LogisticRegression
log_reg = LogisticRegression()
log_reg.fit(X_train, y_train)
log_reg_loss = 1 -  log_reg.score(X_test, y_test)


# tmp_axes =  [-4, 4, -4, 4]
# fig, axes = plt.subplots(1, 3, figsize=(15, 5))
# plt.sca(axes[0])
# plt.title(f'Logistic Regression (loss: {round(log_reg_loss, 3)})')
# plot_predictions(log_reg, tmp_axes)
# plot_dataset(X_test, y_test, tmp_axes)

# plt.sca(axes[1])
# plt.title(f'Logistic KRR (loss: {round(lkrr_loss, 3)})')
# plot_predictions(lkrr, tmp_axes)
# plot_dataset(X_test, y_test, tmp_axes)

# plt.sca(axes[2])
# plt.title(f'Logistic KRGEE (loss: {round(lkrgee_loss, 3)})')
# plot_predictions(lkrgee, tmp_axes)
# plot_dataset(X_test, y_test, tmp_axes)
# plt.tight_layout()



# # # plt.savefig('classification_model_comparison_v0.png', dpi=300)


# # # # ROC curve
from sklearn.metrics import roc_curve, auc
log_reg_probs = log_reg.predict_proba(X_test)[:, 1]
lkrr_probs = lkrr.predict(X_test)
lkrgee_probs = lkrgee.predict(X_test)

log_reg_fpr, log_reg_tpr, _ = roc_curve(y_test, log_reg_probs)
lkrr_fpr, lkrr_tpr, _ = roc_curve(y_test, lkrr_probs)
lkrgee_fpr, lkrgee_tpr, _ = roc_curve(y_test, lkrgee_probs)

log_reg_auc = auc(log_reg_fpr, log_reg_tpr)
lkrr_auc = auc(lkrr_fpr, lkrr_tpr)
lkrgee_auc = auc(lkrgee_fpr, lkrgee_tpr)




print()
print('log_reg_loss: ', log_reg_loss)
print('lkrr_loss: ', lkrr_loss)
print('lkrgee loss: ', lkrgee_loss)
print()

print('log_reg_auc: ', log_reg_auc)
print('lkrr_auc: ', lkrr_auc)
print('lkrgee_auc: ', lkrgee_auc)


tmp_axes =  [-4, 4, -4, 4]
fig, axes = plt.subplots(1, 3, figsize=(15, 5))
plt.sca(axes[0])
plt.title(f'Logistic Regression (loss: {round(log_reg_loss, 3)})')
plot_predictions(log_reg, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)

plt.sca(axes[1])
plt.title(f'Logistic KRR (loss: {round(lkrr_loss, 3)})')
plot_predictions(lkrr, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)

plt.sca(axes[2])
plt.title(f'Logistic KRGEE (loss: {round(lkrgee_loss, 3)})')
plot_predictions(lkrgee, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)
plt.tight_layout()




# plt.figure(figsize=(8, 6))
# plt.plot(log_reg_fpr, log_reg_tpr, 'b-', label='Logistic Regression (area = %0.4f)' % log_reg_auc)
# plt.plot(lkrr_fpr, lkrr_tpr, 'g-', label='Logistic KRR (area = %0.4f)' % lkrr_auc)
# plt.plot(lkrgee_fpr, lkrgee_tpr, 'r-', label='Logistic KRGEE  (area = %0.4f)' % lkrgee_auc)
# plt.plot([0, 1], [0, 1], 'k--')
# plt.xlabel('False Positive Rate')
# plt.ylabel('True Positive Rate')
# plt.legend()
# plt.grid(True)
# plt.tight_layout()








# # from sklearn.metrics import confusion_matrix
# # log_reg_pred = (log_reg_probs >= 0.5).astype(float)
# # lkrr_pred = (lkrr_probs >= 0.5).astype(float)
# # lkrgee_pred = (lkrgee_probs >= 0.5).astype(float)

# # log_reg_cm = confusion_matrix(y_test, log_reg_pred)
# # lkrr_cm = confusion_matrix(y_test, lkrr_pred)
# # lkrgee_cm = confusion_matrix(y_test, lkrgee_pred)

# # import seaborn as sns

# # fig, axes = plt.subplots(1, 3, figsize=(15, 5))
# # plt.sca(axes[0])
# # plt.title('Logistic Regression')
# # plt.xlabel('Predicted')
# # plt.ylabel('True')
# # # # sns.heatmap(log_reg_cm, annot=True, fmt='g', cmap='Blues', cbar=False)

# # # # plt.sca(axes[1])
# # # # plt.title('Logistic KRR')
# # # # plt.xlabel('Predicted')
# # # # plt.ylabel('True')
# # # # sns.heatmap(lkrr_cm, annot=True, fmt='g', cmap='Blues', cbar=False) 

# # # # plt.sca(axes[2])
# # # # plt.title('Logistic KRGEE')
# # # # plt.xlabel('Predicted')
# # # # plt.ylabel('True')
# # # # sns.heatmap(lkrgee_cm, annot=True, fmt='g', cmap='Blues', cbar=False)
# # # # plt.tight_layout()
# # # # plt.savefig('classification_confusion_matrix.png', dpi=300)





# # # # y_pred = (p1(f_test) >= 0.5).astype(float)
# # # # cm = confusion_matrix(y_test, y_pred).astype(float)
# # # # import seaborn as sns
# # # # cm[0,:] = cm[0,:]/np.sum(cm[0,:])
# # # # cm[1,:] = cm[1,:]/np.sum(cm[1,:])

# # # # cm = np.round(cm, 2)
# # # # # cm[1:,1] = cm[1:,1]/np.sum(cm[1:,1])
# # # # sns.heatmap(cm, annot=True, fmt='g', cmap='Blues', cbar=False)
# # # # plt.xlabel('Predicted')
# # # # plt.ylabel('True')
# # # # plt.title('Confusion Matrix')





import numpy as np
import matplotlib.pyplot as plt

from phylokrr.utils import k_fold_cv_random, split_data, PGLS
from phylokrr.datasets import load_vcv, sim_data
from phylokrr.kernels import KRR
np.random.seed(12038)




vcv = load_vcv()
n = vcv.shape[0]
mean_vector = np.zeros(n)
num_test = round(0.40 * n)

# simulate phylogenetically weighted data
X_w, y_w = sim_data(vcv, mean_vector, b=1, n_var=1)


(X_train, X_test,
 y_train, y_test,) = split_data(X_w, y_w, num_test, seed = 12038) # seed defined above



rbf_grid_params = {
    'lambda': np.logspace(-15, 15, 200, dtype=float, base=2),
    'gamma' : np.logspace(-15, 15, 200, dtype=float, base=2),
}

krr_rbf = KRR(kernel='rbf', fit_intercept=True)
pgls_model = PGLS(fit_intercept =  True)

best_params = k_fold_cv_random(X_train, y_train,
                                krr_rbf,
                                rbf_grid_params,
                                verbose = False,
                                folds = 5,
                                sample = 100)

krr_rbf.set_params(**best_params)
krr_rbf.fit(X_train, y_train)
y_pred_cv = krr_rbf.predict(X_test)

pgls_model.fit(X_train, y_train)
y_pred_pgls = pgls_model.predict(X_test)



# write stacked generalization 
# which optimizes the weights such that its sum is 1
# using quadratic programming
import cvxpy as cp
import copy
from phylokrr.utils import P_mat_simple, weight_data, myfunc, P_inv_simple


def stacked_generalization(X, y, covs, num_test = 100, seed = 12038):
    import glob
    path_trees = '/Users/ulises/Desktop/ABL/software/phylokrr/data/r_coal_sim_500spps_100trees'
    vcvs = glob.glob(path_trees + '/*.csv')

    vcv = np.loadtxt(vcvs[0], delimiter=',')
    I_vcv = np.diag(np.diag(vcv))
    covs = [ I_vcv, vcv]

    np.random.seed(12038)

    # n,_ = X.shape
    # noise = np.random.multivariate_normal(np.zeros(n), vcv, 1).reshape(-1,1)
    # X = np.random.multivariate_normal(np.zeros(n), vcv, 1).reshape(-1,1)

    # y = myfunc(X, b=2, type='pol', 
    #         add_noise=False,
    #         noise=noise).reshape(-1,1)
    # y += noise
    # plt.scatter(X, y)

    n,_ = vcv.shape
    noise = np.random.normal(0,1, n).reshape(-1,1)
    X_w = np.random.normal(0,1,n).reshape(-1,1)

    y_w = myfunc(X_w, b=1, type='pol', 
            add_noise=False,
            noise=noise).reshape(-1,1)
    y_w += noise


    P_inv = P_inv_simple(vcv)
    X = P_inv @ X_w
    y = P_inv @ y_w

    subplots = 2

    fig, axs = plt.subplots(1, subplots, figsize=(8, 4))
    axs[0].scatter(  X[:,0], y, color = 'blue',alpha=0.5)
    axs[1].scatter(X_w[:,0], y_w, color = 'red', alpha=0.5)


    params = {
        'lambda': np.logspace(-15, 15, 200, base=2),
        'gamma' : np.logspace(-15, 15, 200,  base=2),
    }


    # take a random sample from range(n
    test_idx = np.random.choice(range(n), size = num_test, replace = False)
    train_idx = list( set(range(n)) - set(test_idx) )

    # X = np.ones(n).reshape(-1,1)
    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    errors = np.zeros((len(test_idx), len(covs)))
    for i,vcv in enumerate(covs):

        # h = 1
        # new_vcv  = np.eye(n) * h + vcv*(1-h)
        # np.max(np.linalg.eigvals(new_vcv))


        vcv_train, vcv_test = vcv[train_idx,:][:,train_idx], vcv[test_idx,:][:,test_idx]
        model = KRR(kernel='rbf', fit_intercept=True)

        X_w_train, y_w_train = weight_data(X_train, y_train, vcv_train)
        X_w_test, y_w_test = weight_data(X_test, y_test, vcv_test)
        # plt.scatter(X_w_train, y_w_train)

        best_params = k_fold_cv_random(X_w_train, y_w_train,
                                        model,
                                        params,
                                        verbose = False,
                                        folds = 5,
                                        sample = 100)

        model.set_params(**best_params)

        model.fit(X_w_train, y_w_train)
        tmp_errors = (y_w_test - model.predict(X_w_test)).ravel()
        
        # normalize errors
        errors[:,i] = tmp_errors /  (max(y_w_test) - min(y_w_test))

        # normalize errors with the first and third interquartile of y_w_test
        # errors[:,i] = tmp_errors / (np.percentile(y_w_test, 75) - np.percentile(y_w_test, 25))
        # errors[:,i] = tmp_errors / np.std(y_w_test)

    Q = errors.T @ errors
    print(Q)
    m = Q.shape[0]
    ones = np.ones(m)

    # Define and solve the CVXPY problem.
    w = cp.Variable(m)
    prob = cp.Problem(cp.Minimize( cp.quad_form(w, Q) ),
                    [ones.T @ w == 1, w >= 0])
    prob.solve()

    # Print result.
    print(np.round(w.value,3))





import numpy as np
import matplotlib.pyplot as plt


from phylokrr.kernels import KRR
from phylokrr.datasets import load_1d_data_example
from phylokrr.utils import weight_data

X_uw_uc,y_uw_uc,vcv = load_1d_data_example(return_cov=True)

# weighting the data by using the variance covariance matrix
X_w, y_w = weight_data(X_uw_uc, y_uw_uc, vcv, use_sd = False)
# second column is the absolute value of the first column
y_w = np.stack([y_w, np.abs(y_w)], axis = 1) 

# plotting data



# # region
fig, axs = plt.subplots(1, 2, figsize=(8, 4))
axs[0].scatter(X_w, y_w[:,0], color = 'blue', alpha=0.5, )
axs[0].set_title('Phylogenetically weighted data')
axs[0].set_xlabel('$x^*$')
axs[0].set_ylabel('$y^*_1$')

axs[1].scatter(X_w, y_w[:,1], color = 'blue', alpha=0.5, )
axs[1].set_title('Phylogenetically weighted data')
axs[1].set_xlabel('$x^*$')
axs[1].set_ylabel('$y^*_2$')
plt.tight_layout()
# endregion

"""## Simple model fitting without Cross-Validation (CV)"""

from phylokrr.utils import split_data

n,_ = X_w.shape
# split data into training and testing sets
num_test = round(0.5*n)

(X_train, X_test,
 y_train, y_test,) = split_data(X_w, y_w, num_test, seed = 12038) # seed defined above


# from phylokrr.kernels import KRR

#region #
from phylokrr.utils import k_fold_cv_random
# set model
pkrr_model = KRR(kernel='rbf', fit_intercept= True)

params = {
    'lambda' : np.logspace(-5, 10, 200, base=2),
    'gamma' : np.logspace(-5, 10, 200,  base=2),
}

best_params = k_fold_cv_random(X_train, y_train,
                                pkrr_model,
                                params,
                                verbose = False,
                                folds = 5,
                                sample = 100)

pkrr_model.set_params(**best_params)

# fit model with phylogenetic covariance matrix
pkrr_model.fit(X_train, y_train,)
# pkrr_model.alpha
y_pred_kernel = pkrr_model.predict(X_test)

"""Let's compare it with the standard phylogenetic regression (i.e., PGLS)"""

from phylokrr.utils import PGLS
# fit standard phylogenetic regression
pgls = PGLS(fit_intercept=True)
pgls.fit(X_train, y_train)
y_pred_pgls = pgls.predict(X_test)
# store figure


subplots =2
fig, axs = plt.subplots(1, subplots, figsize=(9, 4))
axs[0].scatter(X_test, y_test[:,0], color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
axs[0].scatter(X_test, y_pred_kernel[:,0], color = 'green', alpha=0.5, label = 'phyloKRR predictions')
axs[0].scatter(X_test, y_pred_pgls[:,0], color = 'red', alpha=0.5, label = 'PGLS predictions')
axs[0].set_xlabel('$x^*$')
axs[0].set_ylabel('$y^*_1$')
axs[0].legend()
axs[1].scatter(X_test, y_test[:,1], color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
axs[1].scatter(X_test, y_pred_kernel[:,1], color = 'green', alpha=0.5, label = 'phyloKRR predictions')
axs[1].scatter(X_test, y_pred_pgls[:,1], color = 'red', alpha=0.5, label = 'PGLS predictions')
axs[1].set_xlabel('$x^*$')
axs[1].set_ylabel('$y^*_2$')
axs[1].legend()
plt.tight_layout()
# plt.savefig('../../../writings/Kernels/revisions/multioutput.pdf')

r2_kernel = pkrr_model.score(X_test, y_test, metric='r2')
r2_pgls = pgls.score(X_test, y_test, metric='r2')

print(f"R2 for phyloKRR: {r2_kernel}")
print(f"R2 for PGLS: {r2_pgls}")
# endregion ####





# # X_train.shape
# K_train = RBF_kernel(X_train, X_train, pkrr_model.params['gamma'])
# K_test  = RBF_kernel(X_test, X_train,  pkrr_model.params['gamma'])



# cats = 2
# # ones = np.ones((cats,cats))
# ones = np.array([
#     [1, 0], 
#     [1, 1]])
# # repeat K_train in cats by cats blocks
# K_train_aug = np.kron(ones, K_train)
# K_test_aug = np.kron(ones, K_test)

# # stack y_train in a vector colum
# y_train_flat = y_train.flatten('F')
# y_test_flat = y_test.flatten('F')

# pkrr_mlt = KRR(kernel='precomputed', fit_intercept= True)


# pkrr_mlt.params['lambda'] = pkrr_model.params['lambda']
# pkrr_mlt.fit(K_train_aug,y_train_flat)

# y_pred_mlt = pkrr_mlt.predict(K_test_aug)


# subplots =2
# fig, axs = plt.subplots(1, subplots, figsize=(8, 4))
# axs[0].scatter(X_test, y_test_flat[:num_test], color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
# axs[0].scatter(X_test, y_pred_mlt[:num_test], color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
# axs[0].scatter(X_test, y_pred_pgls[:,0], color = 'red', alpha=0.5, label = 'PGLS predictions')
# axs[0].set_xlabel('$x^*$')
# axs[0].set_ylabel('$y^*_1$')
# axs[0].legend()
# axs[1].scatter(X_test, y_test_flat[num_test:], color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
# axs[1].scatter(X_test, y_pred_mlt[num_test:], color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
# axs[1].scatter(X_test, y_pred_pgls[:,1], color = 'red', alpha=0.5, label = 'PGLS predictions')
# axs[1].set_xlabel('$x^*$')
# axs[1].set_ylabel('$y^*_2$')
# axs[1].legend()
# plt.tight_layout()







import numpy as np


import matplotlib.pyplot as plt


from phylokrr.datasets import load_1d_data_example
from phylokrr.utils import weight_data

X_uw_uc,y_uw_uc,vcv = load_1d_data_example(return_cov=True)
X_w, y_w = weight_data(X_uw_uc, y_uw_uc, vcv, use_sd = False)


np.random.seed(12037)


fig, axs = plt.subplots(1, 2, figsize=(8, 4))
axs[0].scatter(X_uw_uc, y_uw_uc, color = 'red', alpha=0.5, )
axs[0].set_title('Unweighted data')
axs[0].set_xlabel('x')
axs[0].set_ylabel('y')
axs[1].scatter(X_w, y_w, color = 'blue', alpha=0.5, )
axs[1].set_title('Phylogenetically weighted data')
axs[1].set_xlabel('$x^*$')
axs[1].set_ylabel('$y^*$')
plt.tight_layout()



from phylokrr.utils import split_data

n,_ = X_w.shape
# split data into training and testing sets 
num_test = round(0.5*n)

(X_train, X_test, 
 y_train, y_test,) = split_data(X_w, y_w, num_test, seed = 12038) # seed defined above



from phylokrr.kernels import KRR
# set model
pkrr_model = KRR(kernel='rbf', fit_intercept= True)
# arbitrarily proposed hyperparameters
params = {'lambda': 1, 'gamma': 1}
# set hyperparamters
pkrr_model.set_params(**params)
# fit model with phylogenetic covariance matrix
pkrr_model.fit(X_train, y_train,)
y_pred_kernel = pkrr_model.predict(X_test)





from phylokrr.utils import PGLS
# fit standard phylogenetic regression
pgls = PGLS(fit_intercept=True)
pgls.fit(X_train, y_train)
y_pred_pgls = pgls.predict(X_test)
# store figure

plt.scatter(X_test, y_test , color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
plt.scatter(X_test, y_pred_kernel, color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
plt.scatter(X_test, y_pred_pgls, color = 'red', alpha=0.5, label = 'PGLS predictions')
plt.xlabel('x')
plt.ylabel('y')
plt.legend()
plt.tight_layout()




from phylokrr.utils import k_fold_cv_random
params = {
    'lambda' : np.logspace(-5, 3, 200, base=2),
    'gamma' : np.logspace(-5, 3, 200,  base=2),
}

best_params = k_fold_cv_random(X_train, y_train,
                                pkrr_model, 
                                params,
                                verbose = False,
                                folds = 3, 
                                sample = 100)



pkrr_model.set_params(**best_params)
pkrr_model.fit(X_train, y_train)
y_pred_cv = pkrr_model.predict(X_test)



fs = 10
plt.scatter(X_test, y_test, color = 'blue' ,
            alpha=0.5, 
            label = 'Testing (unseen) data',)
plt.scatter(X_test, y_pred_cv, 
            color = 'green', alpha=0.5, 
            label = 'phyloKRR predictions',)
plt.scatter(X_test, y_pred_pgls, 
            color = 'red', alpha=0.5, 
            label = 'PGLS predictions',)
plt.xlabel('x', fontsize = fs)
plt.ylabel('y', fontsize = fs)
plt.legend(fontsize = fs)
plt.tight_layout()




r2_kernel = pkrr_model.score(X_test, y_test, metric='r2')
r2_pgls = pgls.score(X_test, y_test, metric='r2')

print(f"R2 for phyloKRR: {r2_kernel}")
print(f"R2 for PGLS: {r2_pgls}")




import numpy as np
from phylokrr.utils import P_mat_simple, myfunc
import pkg_resources


def load_vcv():
    vcv_file = pkg_resources.resource_stream('phylokrr', 'data/test_cov2.csv')
    vcv = np.loadtxt(vcv_file, delimiter=',')
    return vcv


def load_1d_data_example(return_cov=True):
    
    """
    created with the following code:
    
    ```
    import random
    import numpy as np

    from phylokrr.utils import P_inv_simple
    from phylokrr.datasets import load_vcv

    np.random.seed(12037)


    vcv = load_vcv()
    mean_vector = np.zeros(vcv.shape[0])

    X_w_uc = np.random.normal(0, 1, vcv.shape[0])
    # Non-linear response variable (sine curve)
    y_w_uc = np.sin(X_w_uc*1.5).ravel() 

    # Add noise to the response variable
    y_w_uc[::10] += 4 * (0.5 - np.random.rand(X_w_uc.shape[0] // 10))

    # we can attempt to unweight to the original space
    # with the square root of the covariance matrix
    P_inv = P_inv_simple(vcv)
    X_uw_uc, y_uw_uc = P_inv @ X_w_uc, P_inv @ y_w_uc

    Xy = np.stack([X_uw_uc, y_uw_uc], axis=1)
    np.savetxt('../src/data/test_data_unweigthed.csv', Xy, delimiter=',')
    ```

    """

    data_file = pkg_resources.resource_stream('phylokrr', 'data/test_data_unweigthed.csv')
    data = np.loadtxt(data_file, delimiter=',')
    X, y = data[:,0].reshape(-1,1), data[:,1]

    if return_cov:
        return X, y, load_vcv()
    
    else:
        return X, y


def sim_data(vcv, mean_vector, b, 
             on_weighted=True, 
             add_noise = True, 
             noise_var = 1,
             n_var = 3, 
             extra_weights = [], 
             type='pol',
             weight_with_vcv = True,
             vcv2 = None
             ):
    """
    Simulate weighted data for n_var independent variables

    Parameters
    ----------
    vcv: np.array
        covariance matrix where data is simulated from
    
    mean_vector: np.array
        mean vector
    
    b: int
        power of the polynomial function

    on_weighted: bool
        whether to use weighted data or not to obtain the response variable.
        the weights are obtained using the 
        squared root of the inverse of 
        the covariance matrix

    add_noise: bool
        whether to add noise to the response variable

    noise_var: float
        variance of the noise

    extra_weights: list
        weights for the independent variables

    type: str
        type of the function to simulate the data
        if 'sin' then a sine function is used over the
        sum of the independent variables
        if 'pol' then a polynomial function is used over the
        sum of the independent variables
    
    weight_with_vcv: bool
        if true (default) the weighting of the values is 
        done using vcv. Otherwise, vcv2 is used

    vcv2: np.array
        if weight_with_vcv is false, then we use this covariance matrix
        to weight the data
    
    Returns
    -------
    X_w_uc: np.array
        weighted independent variables (uncentered)
    
    y_w_uc: np.array
        weighted response variable (uncentered)
    """
    # n_var = 1
    # mean_vector = np.zeros(vcv.shape[0])
    # b = 3
    n = vcv.shape[0]
    if len(extra_weights):
        assert len(extra_weights) == n_var, "extra weights must have the same length as the number of predictors"
    
    else:
        extra_weights = np.ones(n_var)

    X_uw_uc = np.zeros((n, n_var))
    for j in range(n_var):
        X_uw_uc[:,j] = np.random.multivariate_normal(mean=mean_vector, cov=vcv)*extra_weights[j]
    
    if add_noise:
        noise = np.random.normal(0, noise_var, n)

    if weight_with_vcv:
        P = P_mat_simple(vcv)

    else:
        assert isinstance(vcv2, np.ndarray), 'vcv2 must be a numpy array'
        P = P_mat_simple(vcv2)

    if on_weighted:
        X_w_uc = P @ X_uw_uc
        # the variance and expectation 
        # of the response variable will depend on the moment 
        # that generates b. V(x^b) = E(x^2b) - E(x^b)^2
        # E(x^b) is the expectation in the moment b.
        # Both the expectation and variance are point estimates
        y_w_uc = myfunc(X_w_uc, b=b, type=type, 
                        add_noise=add_noise,
                        noise=noise).ravel()
    else:
        y_uw_uc = myfunc(X_uw_uc, b=b, type=type, 
                         add_noise=add_noise, 
                         noise=noise).ravel()

        X_w_uc = P @ X_uw_uc
        y_w_uc = P @ y_uw_uc

    return X_w_uc, y_w_uc



import numpy as np
import pandas as pd
from scipy.stats import multivariate_normal

# from scipy.special import expit
import statsmodels.api as sm
import matplotlib.pyplot as plt

def plot_dataset(X, y, axes):
    plt.plot(X[:, 0][y==0], X[:, 1][y==0], "bs")
    plt.plot(X[:, 0][y==1], X[:, 1][y==1], "g^")
    plt.axis(axes)
    plt.grid(True, which='major', linestyle='--', linewidth='0.5')
    plt.xlabel(r"$x_1$", fontsize=20)
    plt.ylabel(r"$x_2$", fontsize=20, rotation=0)


# Parameters
n_obs = 1000   # Total observations
rho = 0.5  # Intracluster correlation (exchangeable structure)

# Generate Two Clouds of Points
np.random.seed(10039)

first_half = n_obs // 2

x1 = np.random.normal(loc=1 , scale=1, size=(first_half, 2))  # Spread out around different centers
x2 = np.random.normal(loc=-1, scale=1, size=(n_obs - first_half, 2))  # Spread out around different centers
X = np.concatenate((x1, x2))  # uncorrelated features


vcv = rho * np.ones((n_obs, n_obs)) + (1 - rho) * np.eye(n_obs)  # Exchangeable correlation
E,Q = np.linalg.eig( vcv )
L = Q @ np.diag( E**(1/2) )

X = L @ X  # Now observations are correlated
X = np.real(X)  # Ensure no complex numbers
b = 0
eta = X[:,0]*X[:,1] + b  # Linear predictor
p = 1/(1 + np.exp(-(eta + b)))  # Linear predictor
correlated_errors = multivariate_normal.rvs(mean=np.zeros(n_obs), cov=vcv)

y = p + correlated_errors >= 0.5   # Binary outcome based on x1 and x2
# y = p >= 0.5   # Binary outcome based on x1 and x2
# plot_dataset(X, y, np.array([-1, 1, -1, 1])*5)

# Create DataFrame
df = pd.DataFrame({"x1": X[:,0], "x2": X[:,1], "y": y})

# Fit GEE Model
model = sm.GEE(df["y"], sm.add_constant(df[["x1", "x2"]]), groups=np.ones(n_obs), 
               family=sm.families.Binomial(), cov_struct=sm.cov_struct.Exchangeable())
result = model.fit()

# Display Results
# print(result.summary())

# get pearson residuals
gee_err = np.mean(result.resid_pearson)
print("Pearson residuals: ", gee_err)

y_prob_gee = model.predict(result.params, exog=sm.add_constant(df[["x1", "x2"]]))

y_pred_gee = (y_prob_gee >= .5).astype(float)
loss_gee = np.mean(y_pred_gee != df["y"])  # Accuracy
print('loss GEE: ', loss_gee)



from sklearn.linear_model import LogisticRegression
log_reg = LogisticRegression()
X_l = np.concatenate(( np.ones((X.shape[0], 1)), X), axis=1)  # Add intercept

log_reg.fit(X_l, y)
log_reg_loss = 1 - log_reg.score(X_l, y)
print('log_reg_loss: ', log_reg_loss)
log_reg_probs = log_reg.predict_proba(X_l)[:, 1]  # Get probabilities

# make auc curve
from sklearn.metrics import roc_curve, auc

log_reg_fpr, log_reg_tpr, _ = roc_curve(y, log_reg_probs)
gee_fpr, gee_tpr, _ = roc_curve(y, y_prob_gee)

log_reg_auc = auc(log_reg_fpr, log_reg_tpr)
gee_auc = auc(gee_fpr, gee_tpr)

print('log_reg_auc: ', log_reg_auc)
print('gee_auc: ', gee_auc)

# plt.figure(figsize=(8, 6))
# plt.plot(log_reg_fpr, log_reg_tpr, 'b-', label='Logistic Regression (area = %0.4f)' % log_reg_auc)
# plt.plot(gee_fpr, gee_tpr, 'r-', label='Logistic KRGEE  (area = %0.4f)' % gee_auc)
# plt.plot([0, 1], [0, 1], 'k--')
# plt.xlabel('False Positive Rate')
# plt.ylabel('True Positive Rate')
# plt.legend()
# plt.grid(True)
# plt.tight_layout()




import numpy as np
import random
import numpy as np
import glob
# import matplotlib.pyplot as plt

# seed for reproducibility
np.random.seed(12038)
random.seed(12038)

from phylokrr.datasets import load_vcv,load_1d_data_example
from phylokrr.utils import sim_data, split_data, PGLS, P_mat_simple, myfunc, k_fold_cv_random, P_inv_simple
from phylokrr.kernels import KRR 


rbf_grid_params = {
    'lambda': np.logspace(-10, 5, 200, dtype=float, base=2),
    'gamma': np.logspace(-10, 5, 200, dtype=float, base=2),
}


path_trees = '/Users/ulises/Desktop/ABL/software/phylokrr/data/r_coal_sim_500spps_100trees'
vcvs = glob.glob(path_trees + '/*.csv')

vcv = np.loadtxt(vcvs[0], delimiter=',')
I_vcv = np.diag(np.diag(vcv))
testing_covs = [vcv, I_vcv]


n = vcv.shape[0]
b = 2
n_var = 3
noise_var = 1
sample = 200
num_test = round(0.5*n)


noise = np.random.normal(0, noise_var, n)
X_uw = np.random.normal(0, 1, n).reshape(-1,1)

y_wc = myfunc(X_uw, b=b, type='pol', 
                    add_noise=True,
                    noise=noise).ravel()

P_inv = P_inv_simple(vcv)
X_uw_uc = P_inv @ X_uw
y_uw_uc = P_inv @ y_wc


Y = np.zeros((num_test, len(testing_covs)))
F = np.zeros((num_test, len(testing_covs)))

krr_rbf_tmp = KRR(kernel='rbf', fit_intercept=False)

for i in range(len(testing_covs)):

    vcv_i = testing_covs[i]
    print("Testing cov matrix ", i)
    
    Pi = P_mat_simple(vcv_i)
    # np.sqrt(np.linalg.inv(vcv_i))

    X_wc_tmp = Pi @ X_uw_uc
    y_wc_tmp = Pi @ y_uw_uc
    
    X_wc_tmp = np.ones(X_wc_tmp.shape[0]).reshape(-1,1)

    (X_wc_train_tmp, X_wc_test_tmp,
     y_wc_train_tmp, y_wc_test_tmp,) = split_data(X_wc_tmp, y_wc_tmp, 
                                                  num_test=num_test,
                                                  seed=12038)

    # if i == 0:
    rbf_param_tmp = k_fold_cv_random(X_wc_train_tmp, y_wc_train_tmp,
                                      krr_rbf_tmp, rbf_grid_params, 
                                      verbose=False, folds=2,
                                      sample=sample)
    
    krr_rbf_tmp.set_params(**rbf_param_tmp)
    krr_rbf_tmp.fit(X_wc_train_tmp, y_wc_train_tmp)
    print(krr_rbf_tmp.score(X_wc_test_tmp, y_wc_test_tmp, metric='rmse'))
    out1 = krr_rbf_tmp.predict(X_wc_test_tmp)

    Y[:,i] = y_wc_test_tmp
    F[:,i] = out1

    # # if i == 0:
    # rbf_param_tmp2 = k_fold_cv_random(X_wc_test_tmp,
    #                                    y_wc_test_tmp, 
    #                                    krr_rbf_tmp, 
    #                                    rbf_grid_params, 
    #                                    verbose=False, 
    #                                    folds=2, sample=sample)

    # krr_rbf_tmp.set_params(**rbf_param_tmp2)
    # krr_rbf_tmp.fit(X_wc_test_tmp, y_wc_test_tmp)
    # print(krr_rbf_tmp.score(X_wc_train_tmp, y_wc_train_tmp, metric='rmse'))
    # out2 = krr_rbf_tmp.predict(X_wc_train_tmp)

    # Y[:,i] = np.hstack((y_wc_test_tmp, y_wc_train_tmp))
    # F[:,i] = np.hstack((out1, out2))

# Y -= np.mean(Y, axis=0)
# F -= np.mean(F, axis=0)

# Import packages.
import cvxpy as cp
import numpy as np
# Y = 
Q = (Y - F).T @ (Y - F)
m = Q.shape[0]
ones = np.ones(m)

# Define and solve the CVXPY problem.
w = cp.Variable(m)
prob = cp.Problem(cp.Minimize( cp.quad_form(w, Q) ),
                 [ones.T @ w == 1, w >= 0])
prob.solve()

# Print result.
print(np.round(w.value,3))






from krgee import Logistic, k_fold_cv_random_vcv

import numpy as np
import matplotlib.pyplot as plt


def plot_dataset(X, y, axes):
    plt.plot(X[:, 0][y==0], X[:, 1][y==0], "bs")
    plt.plot(X[:, 0][y==1], X[:, 1][y==1], "g^")
    plt.axis(axes)
    plt.grid(True, which='major', linestyle='--', linewidth='0.5')
    plt.xlabel(r"$x_1$", fontsize=20)
    plt.ylabel(r"$x_2$", fontsize=20, rotation=0)

def plot_predictions(clf, axes):
    x0s = np.linspace(axes[0], axes[1], 100)
    x1s = np.linspace(axes[2], axes[3], 100)
    x0, x1 = np.meshgrid(x0s, x1s)
    X = np.c_[x0.ravel(), x1.ravel()]
    y_pred = clf.predict(X).reshape(x0.shape)
    plt.contourf(x0, x1, y_pred, cmap=plt.cm.bwr , alpha=0.45)


n_updts = 20
lam = .0001
gam = 2.1
tol = 1e-4

import time
from sklearn.datasets import make_moons

X, y = make_moons(n_samples=550,
                noise=0.18,
                random_state=12038,
                shuffle=True)

num_train = int(0.65*X.shape[0])

X_train, y_train = X[:num_train], y[:num_train]
X_test, y_test = X[num_train:], y[num_train:]

vcv_train = np.eye(X_train.shape[0])
vcv_test = np.eye(X_test.shape[0])

lkrr = Logistic(kernel='rbf',
                max_iter=37, lam=lam, 
                verbose=False, tol=tol)

params = {
    'lambda' : np.logspace(-5, 3, 200, base=2),
    'gamma' : np.logspace(-5, 3, 200,  base=2),
}

best_params = k_fold_cv_random_vcv(X_train, y_train, vcv_train, 
                                   lkrr, params,
                                   folds=4, sample=200, 
                                   verbose=True, seed=12038)
print('Best params: ', best_params)
lkrr.set_params(**best_params)



# start = time.time()
lkrr.fit(X_test, y_test, vcv_test)
# print('Time taken: ', time.time() - start)
print('Log. krr loss: ', lkrr.score(X_test, y_test))

from sklearn.linear_model import LogisticRegression
log_reg = LogisticRegression()
log_reg.fit(X_train, y_train)
log_reg_loss = 1 -  log_reg.score(X_test, y_test)
print( 'Log. reg. loss: ', round(log_reg_loss, 2) )


tmp_axes =  [-1.5, 2.5, -1, 1.5]
fig, axes = plt.subplots(1, 2, figsize=(10, 4))
plt.sca(axes[0])
plot_predictions(log_reg, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)

plt.sca(axes[1])
plot_predictions(lkrr, tmp_axes)
plot_dataset(X_test, y_test, tmp_axes)
plt.tight_layout()
# plt.savefig('classification_comparisons.png', dpi=300)




import numpy as np
from phylokrr.utils import split_data

def permutation_importance(X, y, model, num_test = 100, seed = 12038, iterations = 1000):
    """
    Permutation importance

    X: data
    y: target
    model: model
    num_test: number of test samples
    seed: random seed
    iterations: number of iterations

    returns: Feature importance for all features. 
        Each column represents the change in error
    """

    np.random.seed(seed)

    n,p = X.shape

    X_train, X_test, y_train,y_test = split_data(X, y, num_test, seed = seed)
    model.fit(X_train, y_train)
    error_orig = model.score(X_test, y_test)


    out = np.zeros((iterations, p))

    for i in range(iterations):

        for j in range(X_test.shape[1]):
            # Create a copy of X_test
            X_test_copy = X_test.copy()

            # Scramble the values of the given predictor
            X_test_copy[:,j] = np.random.permutation(X_test_copy[:,j])
                        
            # Calculate the new RMSE
            error_perm = model.score(X_test_copy, y_test)

            out[i,j] = error_perm - error_orig

    return out

def sample_feature(X, feature, quantiles, sample, integer = False):
    """
    sample unweighted column
    """
    # X = X0
    # feature =0
    Xpdp = X[:,feature]

    if integer:
        return np.sort(np.unique(Xpdp))
    
    q = np.linspace(
        start = quantiles[0],
        stop  = quantiles[1],
        num   = sample
    )

    # sample from the original feature distribution
    # values are not necessarily found in the original
    # feature column
    Xq = np.quantile(Xpdp, q = q)

    # np.any(Xpdp == Xq[1])
    return Xq    

def pdp_1d(X, feature, model, quantiles = [0,1], sample = 70, integer = False):
    """
    1D partial dependence plot

    feature: feature index
    model: model
    quantiles: quantiles to sample feature values
    sample: number of samples
    integer: boolean to indicate if feature is integer

    returns: Xq, pdp_values
    """

    Xq = sample_feature(X, feature, quantiles, sample, integer)
    pdp_values = []
    for n in Xq:
        # copy original values
        X_tmp = X.copy()
        # make original values with 
        # modified feature column
        X_tmp[:,feature] = n

        pdp_values.append(
            np.mean(
                model.predict(X_tmp)
            )
        )

    out = np.hstack((
        Xq.reshape(-1,1), 
        np.array(pdp_values).reshape(-1,1)
        ))
    
    return out

def pdp_2d(X, f1_idx, f2_idx, model, quantiles = [0, 1], sample = 10, integer = [False, False]):
    """
    2D partial dependence plot

    f1_idx: feature 1 index
    f2_idx: feature 2 index
    model: model
    quantiles: quantiles to sample feature values
    sample: number of samples
    integer: boolean list to indicate if feature is integer


    """

    Xq1 = sample_feature(X, f1_idx, quantiles, sample, integer[0])
    Xq2 = sample_feature(X, f2_idx, quantiles, sample, integer[1])

    n_xq1 = len(Xq1)
    n_xq2 = len(Xq2)

    X1, X2 = np.meshgrid(Xq1, Xq2)

    Y = np.zeros((n_xq1, n_xq2))
    for i in range(n_xq1):
        for j in range(n_xq2):
            # j,i = 0,0
            X_tmp = X.copy()
            X_tmp[:,f1_idx] = X1[i,j]
            X_tmp[:,f2_idx] = X2[i,j]

            Y[i,j] = np.mean( model.predict(X_tmp) )

    return X1, X2, Y




import numpy as np
import matplotlib.pyplot as plt


from phylokrr.datasets import load_vcv
from phylokrr.utils import sim_data, split_data
from phylokrr.inspection import pdp_1d, pdp_2d, permutation_importance



np.random.seed(12037)

vcv = load_vcv()
mean_vector = np.zeros(vcv.shape[0])
b = 2
weights = [1,2,3]



# this function already produces the weighted data
X_w_uc, y_w_uc =  sim_data(vcv, mean_vector, b, extra_weights=weights)



n,_ = X_w_uc.shape
# split data into training and testing sets
num_test = round(0.5*n)

(X_train, X_test,
 y_train, y_test,) = split_data(X_w_uc, y_w_uc, num_test, seed = 12038) # seed defined above




from phylokrr.kernels import KRR
# set model
pkrr_model = KRR(kernel='linear', fit_intercept= True)


from phylokrr.utils import k_fold_cv_random

params = {
    'lambda' : np.logspace(-15, 15, 200, base=2),
    'c'  : np.logspace(-15, 15, 200,  base=2),
}

best_params = k_fold_cv_random(X_train, y_train,
                                pkrr_model,
                                params,
                                verbose = False,
                                folds = 3,
                                sample = 100)

pkrr_model.set_params(**best_params)
pkrr_model.fit(X_train, y_train)

pkrr_model.score(X_test, y_test, metric='r2')



# permutation importance
PFI = permutation_importance(X_test, y_test, pkrr_model, num_test = 50, iterations=100)
np.savetxt('../../../writings/Kernels/revisions/pfi_linear.csv', PFI, delimiter=',', header='Trait1,Trait2,Trait3', comments='')
  

# plot bar chart
# adjust size of the plot
# plt.figure(figsize=(4,3))
# plt.bar(range(3), np.mean(PFI, axis=0), color='gray')
# plt.errorbar(range(3), np.mean(PFI, axis=0), np.std(PFI, axis=0), fmt='.', color='black')
# plt.xticks(range(3), ['Trait 1', 'Trait 2', 'Trait 3'])
# plt.ylabel('RMSE increase')
# plt.title('Permutation Feature Importance')
# plt.tight_layout()
# plt.savefig('../../../writings/Kernels/revisions/pfi_linear.pdf')




pdp_1d_f1 = pdp_1d(X_test, 0, pkrr_model, sample=100)
pdp_1d_f2 = pdp_1d(X_test, 1, pkrr_model, sample=100)
pdp_1d_f3 = pdp_1d(X_test, 2, pkrr_model, sample=100)

np.savetxt('../../../writings/Kernels/revisions/pdp_f1_linear.csv', pdp_1d_f1, delimiter=',', header='Trait1,PDvalue', comments='')
np.savetxt('../../../writings/Kernels/revisions/pdp_f2_linear.csv', pdp_1d_f2, delimiter=',', header='Trait2,PDvalue', comments='')
np.savetxt('../../../writings/Kernels/revisions/pdp_f3_linear.csv', pdp_1d_f3, delimiter=',', header='Trait3,PDvalue', comments='')





# all_pdp = [pdp_1d_f1, pdp_1d_f2, pdp_1d_f3]

# # get the y limits 
# y_lims = []
# for i in range(3):
#     y_lims.append(np.max(all_pdp[i][:,1]))
#     y_lims.append(np.min(all_pdp[i][:,1]))
# y_lims = [np.min(y_lims), np.max(y_lims)]



# fs = 18
# fig, ax = plt.subplots(2,3, figsize=(13/1.5, 7.5/1.5))
# # make a subplot with 1 row and 3 columns
# # fig, ax = plt.subplots(1,3, figsize=(13,5))

# # add a 'A' label to the subplot
# ax[0,0].text(-0.1, 1.1, 'A', transform=ax[0,0].transAxes, 
#              fontsize=fs, fontweight='bold', va='top', ha='right')
# for i in range(3):
#     tmp_pdp = all_pdp[i]
#     ax[0,i].plot(tmp_pdp[:,0], tmp_pdp[:,1])    
#     # ax[i].set_title(f'Trait {i+1}')
#     ax[0,i].set_xlabel(f'Trait {i+1}', size=fs)
#     ax[0,i].set_ylabel('PD value', size=fs)
#     ax[0,i].grid(True)
#     ax[0,i].set_ylim(y_lims)
#     # make space beteween subplots
#     fig.subplots_adjust(wspace=0.3)



# # make space beteween the two plots
# fig.subplots_adjust(hspace=0.3)

# # add a 'B' label to the subplot
# ax[1,0].text(-0.1, 1.1, 'B', transform=ax[1,0].transAxes, 
#              fontsize=fs, fontweight='bold', va='top', ha='right')

# 2d partial dependence plot
feature_combinations = [(0,1), (0,2), (1,2)]
# fig,ax=plt.subplots(1,3, figsize=(13,5))
for i,comb in enumerate(feature_combinations):
    X1, X2, Y = pdp_2d(X_test, comb[0], comb[1], pkrr_model, sample = 50) # 20

    out_file = f'../../../writings/Kernels/revisions/pdp_f{comb[0]}_f{comb[1]}_linear.csv'
    np.savetxt(out_file, np.c_[X1.flatten(), X2.flatten(), Y.flatten()], delimiter=',', header='Trait1,Trait2,PDvalue', comments='')


#     cp = ax[1,i].contourf(X1, X2, Y)
#     fig.colorbar(cp, ax=ax[1,i]) # Add a colorbar to a plot
#     # ax[i].set_title('2D ')
#     ax[1,i].set_xlabel(f'Trait {comb[0] + 1}', size=fs)
#     ax[1,i].set_ylabel(f'Trait {comb[1] + 1}', size=fs)
#     # ax[i].grid(True)
# plt.tight_layout()
# plt.savefig('../../../writings/Kernels/revisions/inspection.pdf')
